import pytest
import tempfile

from calliope.utils import AttrDict

from calliope import analysis

from . import common
from .common import assert_almost_equal, solver


class TestModel:
    @pytest.fixture(scope='module')
    def model(self):
        locations = """
            locations:
                1:
                    level: 0
                    within:
                    techs: ['ccgt', 'demand_electricity']
                    override:
                        ccgt:
                            constraints:
                                e_cap_max: 100
                        demand_electricity:
                            constraints:
                                r: -50
            links:
        """
        config_run = """
            mode: plan
            model: [{techs}, {locations}]
            subset_t: ['2005-01-01', '2005-01-02']
        """
        with tempfile.NamedTemporaryFile() as f:
            f.write(locations.encode('utf-8'))
            f.read()
            model = common.simple_model(config_run=config_run,
                                        config_locations=f.name,
                                        override=AttrDict({'solver': solver}))
        model.run()
        return model

    def test_recompute_levelized_costs(self, model):
        # Cost in solution
        df = model.solution.levelized_cost
        assert_almost_equal(df.at['monetary', 'power', 'total', 'ccgt'], 0.1)
        # Recomputed cost must be the same
        dm = analysis.DummyModel(model.solution)
        recomputed = dm.recompute_levelized_costs('ccgt')
        assert_almost_equal(recomputed['total'], 0.1)

    def test_recompute_levelized_costs_after_changes(self, model):
        # Make changes
        dm = analysis.DummyModel(model.solution)
        dm.config_model.techs.ccgt.costs.monetary.e_cap = 50
        dm.config_model.techs.ccgt.costs.monetary.om_fuel = 1.0
        # Recomputed cost
        recomputed = dm.recompute_levelized_costs('ccgt')
        assert_almost_equal(recomputed['total'], 1.0, tolerance=0.001)



import pytest

import pandas as pd

import calliope


class TestMaskWhereZero:
    @pytest.fixture(scope='module')
    def testdata(self):
        d = calliope.utils.AttrDict()
        df1 = pd.DataFrame({'a': list(range(1, 11)),
                            'b': list(range(20, 30))})
        df2 = pd.DataFrame({'a': [0, 0, 0, 0, 1, 2, 3, 4, 0, 0],
                            'b': [0, 0, 0, 1, 2, 1, 1, 2, 0, 0]})
        df3 = pd.DataFrame({'a': [0, 0, 0, 0, 1, 2, 3, 4, 1, 0],
                            'b': [0, 0, 0, 1, 2, 1, 1, 2, 0, 0]})
        df4 = pd.DataFrame({'a': [0, 0, 0, 0, 1, 2, 3, 4, 1, 2],
                            'b': [0, 0, 0, 1, 2, 1, 1, 2, 0, 0]})
        df5 = pd.DataFrame({'a': [1, 2, 2, 3, 1, 2, 0, 0, 0, 0],
                            'b': [0, 0, 0, 1, 2, 1, 1, 0, 0, 0]})
        d.set_key('r.tech1', df1)
        d.set_key('r.tech2', df2)
        d.set_key('r.tech3', df3)
        d.set_key('r.tech4', df4)
        d.set_key('r.tech5', df5)
        d.set_key('e_eff.tech1', df1 / 100.0)
        return d

    def test_mask_zero_no_zeros(self, testdata):
        data = testdata
        mask = calliope.time_masks.mask_zero(data, 'tech1')
        assert mask.tolist() == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    def test_mask_zero_start_and_end_zero(self, testdata):
        data = testdata
        mask = calliope.time_masks.mask_zero(data, 'tech2')
        assert mask.tolist() == [1, 1, 1, 0, 0, 0, 0, 0, 1, 1]
        res = calliope.time_tools.masks_to_resolution_series([mask])
        assert res.tolist() == [3, -1, -1, 0, 0, 0, 0, 0, 2, -1]

    def test_mask_zero_one_zero(self, testdata):
        data = testdata
        mask = calliope.time_masks.mask_zero(data, 'tech3')
        assert mask.tolist() == [1, 1, 1, 0, 0, 0, 0, 0, 0, 1]
        res = calliope.time_tools.masks_to_resolution_series([mask])
        assert res.tolist() == [3, -1, -1, 0, 0, 0, 0, 0, 0, 0]

    def test_mask_zero_start_zero(self, testdata):
        data = testdata
        mask = calliope.time_masks.mask_zero(data, 'tech4')
        assert mask.tolist() == [1, 1, 1, 0, 0, 0, 0, 0, 0, 0]
        res = calliope.time_tools.masks_to_resolution_series([mask])
        assert res.tolist() == [3, -1, -1, 0, 0, 0, 0, 0, 0, 0]

    def test_mask_zero_end_zero(self, testdata):
        data = testdata
        mask = calliope.time_masks.mask_zero(data, 'tech5')
        assert mask.tolist() == [0, 0, 0, 0, 0, 0, 0, 1, 1, 1]
        res = calliope.time_tools.masks_to_resolution_series([mask])
        assert res.tolist() == [0, 0, 0, 0, 0, 0, 0, 3, -1, -1]


# class TestDownsampling:
#     def test_infinity_test_no_inf(self):
#         pass

#     def test_infinity_test_all_inf(self):
#         pass

#     def test_infinity_test_mixed_inf(self):
#         pass

#     def test_reduce_average(self, df):
#         pass

#     def test_reduce_sum(self, df):
#         pass

#     def test_reduce_cut(self, df):
#         pass



import pytest
import tempfile

from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver


class TestModel:
    @pytest.fixture(scope='module')
    def model(self):
        locations = """
            locations:
                1:
                    level: 0
                    within:
                    techs: ['demand_electricity']
                    override:
                        demand_electricity:
                            constraints:
                                r: -10
                sub1:
                    level: 1
                    within: 1
                    techs: ['ccgt']
                    override:
                        ccgt:
                            constraints:
                                e_cap_max: 9
                sub2:
                    level: 1
                    within: 1
                    techs: ['pv']
                    override:
                        pv:
                            x_map: 'sub2: demand'
                            constraints:
                                e_cap_max: 9
                                r: file=demand-sin_r.csv
                                r_scale_to_peak: 5

            links:
        """
        config_run = """
            mode: plan
            model: [{techs}, {locations}]
            subset_t: ['2005-01-01', '2005-01-02']
        """
        with tempfile.NamedTemporaryFile() as f:
            f.write(locations.encode('utf-8'))
            f.read()
            model = common.simple_model(config_run=config_run,
                                        config_locations=f.name,
                                        override=AttrDict({'solver': solver}))
        model.run()
        return model

    def test_model_solves(self, model):
         assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_balanced(self, model):
        df = model.solution.node
        assert_almost_equal(df.loc['e:power', 'pv', :, :].iloc[0, :].sum(), 7.5,
                            tolerance=0.01)
        assert (df.loc['e:power', 'pv', :, :].sum(1) +
                df.loc['e:power', 'ccgt', :, :].sum(1) ==
                -1 * df.loc['e:power', 'demand_electricity', :, :].sum(1)).all()

    def test_model_costs(self, model):
        df = model.solution.levelized_cost
        assert_almost_equal(df.at['monetary', 'power', 'total', 'ccgt'], 0.1,
                            tolerance=0.001)



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

analysis.py
~~~~~~~~~~~

Functionality to analyze model results.

"""

import itertools
import logging

try:
    import matplotlib.pyplot as plt
    from matplotlib.colors import ListedColormap
except ImportError:
    logging.debug('Matplotlib could not be imported, '
                  'no plotting will be available.')
import numpy as np
import pandas as pd

from . import utils
from . import analysis_utils as au


def plot_carrier_production(solution, carrier='power', subset_t=None,
                            **kwargs):
    """
    Generate a stackplot of the production by the given ``carrier``.

    Parameters
    ----------
    solution : solution container
    carrier : str, default 'power'
    subset_t : 2-tuple of str or None, default None
        Specify a date subset for which to plot, for example,
        ('2005-02-01', '2005-02-10') , or ('2005-02-01', None)
    **kwargs : passed to ``plot_timeseries`` (see its documentation)

    """
    data = solution.node['e:{}'.format(carrier)].sum(axis=2)
    if subset_t:
        data = data.loc[slice(*subset_t), :]
    plot_timeseries(solution, data, carrier=carrier, **kwargs)


def plot_timeseries(solution, data, carrier='power', demand='demand_power',
                    types=['supply', 'conversion', 'storage', 'unmet_demand'],
                    colormap=None, ticks=None):
    """
    Generate a stackplot of the ``data`` for the given ``carrier``,
    plotting the ``demand`` on top.

    Use ``plot_carrier_production`` for a simpler way to plot production by
    a given carrier.

    Parameters
    ----------
    solution : solution container
    data : pandas DataFrame
        subset of solution to plot
    carrier : str, default 'power'
        name of the carrier to plot
    demand : str, default 'demand_power'
        name of a demand tech whose time series to plot on top
    types : list, default ['supply', 'conversion', 'storage', 'unmet_demand']
        Technology types to include in the plot.
    colormap : matplotlib colormap
        Colormap to use, if not given, the colors specified for each
        technology in the solution's metadata are used.
    ticks : str, default None
        Where to draw x-axis (time axis) ticks. By default (None),
        auto-detects, but can manually set to either 'hourly', 'daily',
        or 'monthly'.

    """
    # Determine ticks
    if not ticks:
        timespan = (data.index[-1] - data.index[0]).days
        if timespan <= 2:
            ticks = 'hourly'
        elif timespan < 14:
            ticks = 'daily'
        else:
            ticks = 'monthly'
    # Set up time series to plot, dividing it by time_res_series
    time_res = solution.time_res
    plot_df = data.divide(time_res, axis='index')
    # Get tech stack and names
    df = solution.metadata[solution.metadata.carrier == carrier]
    query_string = au._get_query_string(types)
    stacked_techs = df.query(query_string).index.tolist()
    # Put stack in order according to stack_weights
    weighted = df.stack_weight.order(ascending=False).index.tolist()
    stacked_techs = [y for y in weighted if y in stacked_techs]
    names = [df.at[y, 'name'] for y in stacked_techs]
    # If no colormap given, derive one from colors given in metadata
    if not colormap:
        colors = [solution.metadata.at[i, 'color'] for i in stacked_techs]
        colormap = ListedColormap(colors)
    # Plot!
    ax = au.stack_plot(plot_df, stacked_techs, colormap=colormap,
                       alpha=0.9, ticks=ticks, legend='right', names=names)
    ax.plot(plot_df[demand].index,
            plot_df[demand] * -1,
            color='black', lw=1, ls='-')
    return ax


def plot_installed_capacities(solution,
                              types=['supply', 'conversion', 'storage'],
                              unit_multiplier=1.0,
                              unit_label='kW',
                              **kwargs):
    """
    Plot installed capacities (``e_cap``) with a bar plot.

    Parameters
    ----------
    types : list, default ['supply', 'conversion', 'storage']
        Technology types to include in the plot.
    unit_multiplier : float or int, default 1.0
        Multiply installed capacities by this value for plotting.
    unit_label : str, default 'kW'
        Label for capacity values, adjust this when
        changing ``unit_multiplier``.
    **kwargs : are passed to ``pandas.DataFrame.plot()``

    """
    query_string = au._get_query_string(types)
    supply_cap = solution.metadata.query(query_string).index.tolist()

    df = solution.parameters.e_cap.loc[:, supply_cap]

    weighted = solution.metadata.stack_weight.order(ascending=False).index.tolist()
    stacked_techs = [y for y in weighted if y in df.columns]

    df = df.loc[:, stacked_techs] * unit_multiplier

    names = [solution.metadata.at[y, 'name'] for y in df.columns]
    colors = [solution.metadata.at[i, 'color'] for i in df.columns]
    colormap = ListedColormap(colors)
    proxies = [plt.Rectangle((0, 0), 1, 1, fc=i)
               for i in colors]

    # Order the locations nicely, but only take those locations that actually
    # exists in the current solution
    if ('metadata' in solution.config_model and
            'location_ordering' in solution.config_model.metadata):
        meta_config = solution.config_model.metadata
        for index, item in enumerate(meta_config.location_ordering):
            if item in df.index:
                df.at[item, 'ordering'] = index
        df = df.sort('ordering', ascending=False)
        df = df.drop('ordering', axis=1)

    ax = df.plot(kind='barh', stacked=True, legend=False, colormap=colormap,
                 **kwargs)
    leg = au.legend_on_right(ax, style='custom', artists=proxies, labels=names)

    ylab = ax.set_ylabel('')
    xlab = ax.set_xlabel('Installed capacity ({})'.format(unit_label))

    return ax


def plot_transmission(solution, tech='hvac', carrier='power',
                      labels='utilization',
                      figsize=(15, 15), fontsize=9):
    """
    Plot transmission links on a map. Requires that model metadata have
    been defined with a lat/lon for each model location and a boundary for
    the map display.

    Requires Basemap and NetworkX to be installed.

    Parameters
    ----------
    solution : solution container
    tech : str, default 'hvac'
        Which transmission technology to plot.
    carrier : str, default 'power'
        Which carrier to plot transmission for.
    labels : str, default 'utilization'
        Determines how transmission links are labeled, either
        `transmission` or `utilization`.
    figsize : (int, int), default (15, 15)
        Size of resulting figure.
    fontsize : int, default 9
        Font size of figure labels.

    """
    import networkx as nx

    # Determine maximum that could have been transmitted across a link
    def get_edge_capacity(solution, a, b):
        hrs = solution.time_res.sum()
        cap = solution.parameters.at['e_cap_net', a, '{}:'.format(tech) + b] * hrs
        return cap

    # Get annual power transmission between zones
    zones = sorted(solution.node.minor_axis.tolist())
    trans_tech = lambda x: '{}:{}'.format(tech, x)
    df = pd.DataFrame({zone: solution.node.loc['e:{}'.format(carrier),
                                               trans_tech(zone), :, :].sum()
                      for zone in zones})

    # Set smaller than zero to zero --
    # we only want to know about 'production' from
    # transmission, not their consumptions
    df[df < 0] = 0

    # Create directed graph
    G = nx.from_numpy_matrix(df.as_matrix().T, create_using=nx.DiGraph())
    G = nx.relabel_nodes(G, dict(list(zip(list(range(len(zones))), zones))))

    # Transmission
    edge_transmission = {edge: int(round(df.at[edge[1], edge[0]] / 1e6))
                         for edge in G.edges()}

    # Utilization ratio
    edge_use = {(a, b): (df.at[a, b] + df.at[b, a])
                / get_edge_capacity(solution, a, b)
                for (a, b) in G.edges()}

    # Set edge labels
    if labels == 'utilization':
        edge_labels = {k: '{:.2f}'.format(v) for k, v in edge_use.items()}
    elif labels == 'transmission':
        edge_labels = edge_transmission

    # Set edge colors
    edge_colors = [edge_use[i] for i in G.edges()]

    ax, m = au.plot_graph_on_map(solution.config_model,
                                 edge_colors, edge_labels,
                                 figsize, fontsize)

    return ax


def get_delivered_cost(solution, cost_class='monetary', carrier='power',
                       count_unmet_demand=False, unit_multiplier=1.0):
    """
    Get the levelized cost per unit of energy delivered for the given
    ``cost_class`` and ``carrier``.

    Parameters
    ----------
    solution : solution container
    cost_class : str, default 'monetary'
    carrier : str, default 'power'
    count_unmet_demand : bool, default False
        Whether to count the cost of unmet demand in the final
        delivered cost
    unit_multiplier : float or int, default 1.0
        Adjust unit of the returned cost value. For example, if model units
        are kW and kWh, ``unit_multiplier=1.0`` will return cost per kWh, and
        ``unit_multiplier=0.001`` will return cost per MWh.

    """
    summary = solution.summary
    meta = solution.metadata
    carrier_subset = meta[meta.carrier == carrier].index.tolist()
    if count_unmet_demand is False:
        carrier_subset.remove('unmet_demand_' + carrier)
    cost = solution.costs.loc[cost_class, :, carrier_subset].sum().sum()
    # Actually, met_demand also includes demand "met" by unmet_demand
    met_demand = summary.at['demand_' + carrier, 'e_con']
    try:
        unmet_demand = summary.at['unmet_demand_' + carrier, 'e_con']
    except KeyError:
        unmet_demand = 0
    if count_unmet_demand is False:
        demand = met_demand + unmet_demand  # unmet_demand is positive, add it
    else:
        demand = met_demand

    return cost / demand * -1


def get_group_share(solution, techs, group_type='supply',
                    var='e_prod'):
    """
    From ``solution.summary``, get the share of the given list of ``techs``
    from the total for the given ``group_type``, for the given ``var``.

    """
    summary = solution.summary
    meta = solution.metadata
    group = meta.query('type == "' + group_type + '"').index.tolist()
    supply_total = summary.loc[group, var].sum()
    supply_group = summary.loc[techs, var].sum()
    return supply_group / supply_total


def get_supply_groups(solution):
    """
    Get individual supply technologies and those groups that define
    group == True, for purposes of calculating diversity of supply

    """
    # idx_1: group is True and '|' in members
    grp_1 = solution.shares.query('group == True & type == "supply"')
    idx_1 = grp_1[(grp_1.members != grp_1.index)
                  & (grp_1.members.str.contains('\|'))].index.tolist()
    # idx_2: group is False and no '|' in members
    grp_2 = solution.shares.query('group == False & type == "supply"')
    idx_2 = grp_2[grp_2.members == grp_2.index].index.tolist()
    # Also drop entries from idx_2 that are already covered by
    # groups in idx_1
    covered = [i.split('|')
               for i in solution.shares.loc[idx_1, 'members'].tolist()]
    covered_flat = [i for i in itertools.chain.from_iterable(covered)]
    idx_2 = [i for i in idx_2 if i not in covered_flat]
    return idx_1 + idx_2


def get_unmet_demand_hours(solution, carrier='power', details=False):
    """
    Get information about unmet demand from ``solution``.

    Parameters
    ----------
    solution : solution container
    carrier : str, default 'power'
    details : bool, default False
        By default, only the number of hours with unmet are returned. If
        details is True, a dict with 'hours', 'timesteps', and 'dates' keys
        is returned instead.

    """
    unmet = solution.node['e:' + carrier]['unmet_demand_' + carrier].sum(1)
    timesteps = len(unmet[unmet > 0])
    hours = solution.time_res[unmet > 0].sum()
    if details:
        return {'hours': hours, 'timesteps': timesteps,
                'dates': unmet[unmet > 0].index}
    else:
        return hours


def areas_below_resolution(solution, resolution):
    """
    Returns a list of (start, end) timestamp tuples delimiting those
    areas in the solution below the given timestep resolution (in hours).

    """
    selected = solution.time_res[solution.time_res < resolution]
    return list(au._get_ranges(selected.index.tolist()))


def get_swi(solution, shares_var='e_cap', exclude_patterns=['unmet_demand']):
    """
    Returns the Shannon-Wiener diversity index.

    :math:`SWI = -1 \\times \sum_{i=1}^{I} p_{i} \\times \ln(p_{i})`

    where where I is the number of categories and :math:`p_{i}`
    is each category's share of the total (between 0 and 1).

    :math:`SWI` is zero when there is perfect concentration.

    """
    techs = get_supply_groups(solution)
    for pattern in exclude_patterns:
        techs = [t for t in techs if pattern not in t]
    swi = -1 * sum((p * np.log(p))
                   for p in [solution.shares.at[y, shares_var] for y in techs]
                   if p > 0)
    return swi


def get_hhi(solution, shares_var='e_cap', exclude_patterns=['unmet_demand']):
    """
    Returns the Herfindahl-Hirschmann diversity index.

    :math:`HHI = \sum_{i=1}^{I} p_{i}^2`

    where :math:`p_{i}` is the percentage share of each technology i (0-100).

    :math:`HHI` ranges between 0 and 10,000. A value above 1800 is
    considered a sign of a concentrated market.

    """
    techs = get_supply_groups(solution)
    for pattern in exclude_patterns:
        techs = [t for t in techs if pattern not in t]
    hhi = sum((solution.shares.at[y, shares_var] * 100.) ** 2 for y in techs)
    return hhi


def get_domestic_supply_index(solution):
    """
    Assuming that ``solution`` specifies a ``domestic`` cost class to
    give each technology a domesticity score, return the total domestic
    supply index for the given solution.

    """
    idx = solution.metadata.query('type == "supply"').index.tolist()
    dom = (solution.costs.domestic.loc[:, idx].sum().sum() /
           solution.totals.loc['power', 'ec_prod', :, :].sum().sum())
    return dom


class DummyModel(object):
    def __init__(self, solution):
        """
        Create a dummy model object from a model solution,
        which gives direct access to the ``config_model`` AttrDict and
        the ``get_option``, ``get_cost``, and ``get_depreciation``
        methods.

        Furthermore, it provides several ``recompute_`` methods
        to recalculate costs for a given technology after modifying the
        underlying data.

        """
        self.config_model = solution.config_model
        self.solution = solution
        data_keys = [k for k in solution.keys() if 'data/' in k]
        self.data = utils.AttrDict()
        for k in data_keys:
            # Turn 'data/r/pv' into 'r/pv', then turn 'r/pv' into 'r.pv'
            data_k = k.split('/', 1)[1] \
                      .replace('/', '.')
            self.data.set_key(data_k, solution[k])

        # Add getters
        self._get_option = utils.option_getter(self.config_model, solution)
        self._get_cost = utils.cost_getter(self._get_option)
        self._get_dep = utils.depreciation_getter(self._get_option)

    def get_option(self, *args, **kwargs):
        return self._get_option(*args, **kwargs)

    def get_cost(self, *args, **kwargs):
        return self._get_cost(*args, **kwargs)

    def get_depreciation(self, *args, **kwargs):
        return self._get_dep(*args, **kwargs)

    def recompute_investment_costs(self, y, k='monetary',
                                   cost_adjustment=1.0):
        """
        Recompute investment costs for the given technology ``y``.

        ``cost_adjustment`` allows linear scaling of the costs by the
        given amount.

        """
        get_cost = self.get_cost
        solution = self.solution
        cost_con = pd.DataFrame()
        cost_con['e_cap'] = (get_cost('e_cap', y, k)
                             * solution.parameters['e_cap'][y])
        cost_con['r_cap'] = (get_cost('r_cap', y, k)
                             * solution.parameters['r_cap'][y])
        cost_con['r_area'] = (get_cost('r_area', y, k)
                              * solution.parameters['r_area'][y])
        cost_con['rb_cap'] = (get_cost('rb_cap', y, k)
                              * solution.parameters['rb_cap'][y])
        cost_con['s_cap'] = (get_cost('s_cap', y, k)
                             * solution.parameters['s_cap'][y])
        cost_con['total'] = cost_con.sum(axis=1)
        cost_con['total_per_e_cap'] = (cost_con['total']
                                       / solution.parameters['e_cap'][y])
        cost_con.at['total', 'total'] = cost_con['total'].sum()
        total_per_e_cap = (cost_con.at['total', 'total']
                           / solution.parameters['e_cap'][y].sum())
        cost_con.at['total', 'total_per_e_cap'] = total_per_e_cap
        return cost_con * cost_adjustment

    def recompute_operational_costs(self, y, k='monetary', carrier='power',
                                    cost_adjustment=1.0):
        """
        Recompute operational costs for the given technology ``y``.

        ``cost_adjustment`` lineary scales the construction costs used
        internally, but is not applied to the rest of the operational
        costs.

        """
        get_cost = self.get_cost
        get_depreciation = self.get_depreciation
        solution = self.solution
        # Here we want gross production
        production = solution.totals[carrier].es_prod[y]
        fuel = (solution.node.loc['rs', y, :, :].sum()
                / self.get_option(y + '.constraints.r_eff'))
        fuel_rb = (solution.node.loc['rbs', y, :, :].sum()
                   / self.get_option(y + '.constraints.rb_eff'))
        icf = self.recompute_investment_costs
        cost_con = icf(y, k=k, cost_adjustment=cost_adjustment)['total'].sum()
        cost_op = pd.DataFrame()
        cost_op['om_frac'] = (get_cost('om_frac', y, k) * cost_con
                              * get_depreciation(y, k)
                              * (solution.time_res.sum() / 8760))
        cost_op['om_fixed'] = (get_cost('om_fixed', y, k)
                               * solution.parameters['e_cap'][y]
                               * (sum(solution.time_res) / 8760))
        cost_op['om_var'] = get_cost('om_var', y, k) * production
        cost_op['om_fuel'] = get_cost('om_fuel', y, k) * fuel
        cost_op['om_rb'] = get_cost('om_rb', y, k) * fuel_rb
        cost_op['total'] = cost_op.sum(axis=1)
        return cost_op

    def recompute_levelized_costs(self, y, k='monetary', carrier='power',
                                  cost_adjustment=1.0):
        """
        Recompute levelized costs for the given technology ``y``.

        ``cost_adjustment`` lineary scales the construction costs used
        internally to compute levelized costs.

        """
        get_depreciation = self.get_depreciation
        solution = self.solution
        icf = self.recompute_investment_costs
        cost_con = icf(y, k=k, cost_adjustment=cost_adjustment)['total']
        ocf = self.recompute_operational_costs
        cost_op = ocf(y, k=k, carrier=carrier,
                      cost_adjustment=cost_adjustment)['total']
        # Here we want net production
        production = solution.totals[carrier].ec_prod[y]
        lc = (get_depreciation(y, k) * (solution.time_res.sum() / 8760)
              * cost_con + cost_op) / production
        lc.at['total'] = (get_depreciation(y, k)
                          * (solution.time_res.sum() / 8760)
                          * cost_con.total
                          # * cost_con.drop(['total'], axis=0).sum()
                          + cost_op.sum()) / production.sum()
        return lc



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

time_tools.py
~~~~~~~~~~~~~

Provides the TimeSummarizer class to dynamically adjust time resolution.

Also provides the masks_to_resolution_series function. A resolution series
is a pandas Series with the same index as the input data, where each
time step is one of:

* a positive integer (signifying how many timesteps to summarize),
* -1 (following a positive integer and marking the timesteps that
  are summarized),
* or 0 (no resolution adjustment to this timestep).

The name of a resolution series must always be 'resolution_series'.

"""

import logging

import numpy as np
import pandas as pd

from . import utils


class TimeSummarizer(object):
    """
    Provides methods to reduce time resolution for the given model data.

    """
    def __init__(self):
        super(TimeSummarizer, self).__init__()
        # Format: {'data item': 'method'}
        self.known_data_types = {'_dt': 'NOOP',
                                 'a': self._reduce_r_to_a,
                                 'r': self._reduce_sum,
                                 'r_eff': self._reduce_average,
                                 'e_eff': self._reduce_average}
        self.source_parameters = {'a': 'r'}

    def _apply_method(self, data, method, s, param, src_param=None,
                      subkey=None):
        if method == 'NOOP':
            # NOOP method means we don't do anything, so return straight away
            return
        i = data._dt.index[s.start]
        if not src_param:
            src_param = param
        if subkey:
            data[param][subkey].loc[i, :] = method(data[src_param][subkey][s])
        else:
            data[param].at[i] = method(data[src_param][s])

    def _reduce_resolution(self, data, resolution, t_range):
        """
        Helper function called by dynamic_timestepper.

        Parameters
        ----------
        data : AttrDict
            data container
        resolution : int
            new time step resolution (in hours)
        t_range : (int, int)
            (absolute) time indices over which to apply resolution reduction

        """
        # Initialize some common data
        resolution = resolution / data.time_res_data
        assert resolution.is_integer()
        resolution = int(resolution)
        self.resolution = resolution
        # Set up time range slice
        s = slice(*t_range)
        # Go through each item in data and apply the appropriate method to it
        for k in list(data.keys()):
            if k in list(self.known_data_types.keys()):
                how = self.known_data_types[k]
                if k in self.source_parameters:
                    src = self.source_parameters[k]
                else:
                    src = None
                if isinstance(data[k], utils.AttrDict):
                    for kk in list(data[k].keys()):
                        self._apply_method(data, how, s, param=k,
                                           src_param=src, subkey=kk)
                else:
                    self._apply_method(data, how, s, param=k,
                                       src_param=src)
            else:  # k is not in known_data_types.keys()
                msg = 'Encountered unknown data type, skipping: {}'.format(k)
                logging.debug(msg)

    def dynamic_timestepper(self, data, series):
        """``series`` must be a series with the same length as the given data.

        For each timestep, the ``series`` can either be 0 (no
        adjustment), >0 (marks the first timestep to be compressed, with
        the integer value giving the number of timesteps to compress) or
        -1 (marks timesteps following a >0 timestep that will be compressed).

        For example, ``series`` could contain::

            [0, 0, 0, 3, -1, -1, 0, 0, 2, -1, 2, -1]

        This would result in data with the following new timestep resolution::

            [1, 1, 1, 3, 1, 1, 2, 2]

        Warning: modifies the passed data object in-place.
        Returns None on success.

        """
        assert len(series) == len(data['_dt']), \
            'Mask and data must have same length.'
        df = pd.DataFrame(series, index=series.index)
        df.columns = ['summarize']  # rename the single column
        df['time_res'] = data.time_res_static
        df['to_keep'] = 1
        # Get all time steps that need summarizing
        entry_points = df.summarize[df.summarize > 0]
        # 1. Summarize by calling _reduce_resolution for the desired ranges
        # This step still keeps the summarized data but marks it by setting
        # `to_keep` 0
        for i, v in entry_points.items():
            ifrom = i
            ito = i + int(v / data.time_res_static)
            resolution = v
            self._reduce_resolution(data, resolution, t_range=(ifrom, ito))
            # Mark the rows that need to be killed with 0
            # Need ifrom+1 and ito-1 because .loc includes start and end slice
            df.loc[ifrom + 1:ito - 1, 'to_keep'] = 0
            df.at[ifrom, 'time_res'] = resolution
        # 2. Replace all data with its subset where to_keep is 1
        # Create boolean mask with the right length -- but ignore index by
        # turning it into a list (since data is indexed by timestep id)
        boolean_mask = (df.to_keep == 1).tolist()
        for k in list(data.keys()):
            if k in list(self.known_data_types.keys()):
                if isinstance(data[k], utils.AttrDict):
                    for kk in list(data[k].keys()):
                        data[k][kk] = data[k][kk][boolean_mask]
                else:
                    data[k] = data[k][boolean_mask]
            # NB unknown data types are checked for and logged earlier, inside
            # _reduce_resolution()
        data.time_res_series = df.time_res[boolean_mask]
        # `df` is zero-indeed, so we have to re-set the actual timesteps as
        # index for time_res_series
        data.time_res_series.index = data._dt.index
        # Update data.time_res_static if the resolution reduction was uniform
        if len(data.time_res_series.unique()) == 1:
            data.time_res_static = resolution
        data.time_res_native = 0  # Unset native time_res flag

    def reduce_resolution(self, data, resolution):
        """
        Reduces the resolution of the entire ``data`` to ``resolution``,
        which must be an integer and greater than the initial resolution.

        Modifies the passed data object in-place and returns None.

        """
        series = resolution_series_uniform(data, resolution)
        self.dynamic_timestepper(data, series)

    def _reduce_average(self, df):
        return df.mean(0)

    def _reduce_sum(self, df):
        return df.sum(0)

    def _reduce_r_to_a(self, df):
        """
        Based on resource data given in ``df``, generate availability data

        """
        if (df.sum(axis=0) == np.inf).all():
            # If the df is all infinite, availability is 1
            result = 1
        elif df.max().max() == 0 and df.min().min() == 0:
            # And if it's all zero, availability is 0
            result = 0
        else:
            result = df.sum() / (df.max() * len(df))
            result = result.fillna(0)  # Fill NaNs from division by zero
        return result


def masks_to_resolution_series(masks, how='or', resolution=None):
    """
    Converts a list of overlapping masks into a series of time step
    resolutions.

    Parameters
    ----------
    how : str, default 'or'
        ``or`` or ``and``.
    resolution : int, default None
        If given, will break up masked areas into timesteps of at most
        the given length (with possibly a over timestep at a lower
        length at the end of the masked area). If None, all contingent
        masked areas become single timesteps.

    """
    if not isinstance(masks, list) or isinstance(masks, tuple):
        masks = [masks]
    # combine all masks into one
    df = pd.DataFrame({i: x for i, x in enumerate(masks)})
    if how == 'or':
        mask = df.sum(axis=1)
    elif how == 'and':
        # joiner: only return 1 if all items in the row are 1, else return 0
        joiner = lambda row: 1 if sum(row) == len(row) else 0
        mask = df.apply(joiner, axis=1)
    istart = 0
    end = False
    while not end:
        ifrom = mask[istart:].argmax()
        ito = mask[ifrom:].argmin()
        if ifrom == ito:  # Reached the end!
            ito = len(mask)
            end = True
            # If `summarize` is zero at the very last entry
            # (`ito - `), we break out of the
            # loop to prevent it from adding a spurious summarization
            if mask[ito - 1] == 0:
                break
        step_resolution = ito - ifrom
        mask[ifrom] = step_resolution
        mask[ifrom + 1:ito] = -1
        # Correct edge case where only one timestep would be "summarized"
        if mask[ifrom] >= 1 and step_resolution == 1:
            mask[ifrom] = 0
        istart = ito
    # Apply max_timesteps
    if resolution:
        for index, value in mask[mask > resolution].iteritems():
            end_index = index + value
            summary_index = list(range(index, end_index, resolution))
            for i in summary_index:
                if i + resolution < end_index:
                    mask[i] = resolution
                else:  # Make sure the last timestep isn't too long
                    mask[i] = end_index - i
    mask.name = 'resolution_series'
    return mask


def resolution_series_uniform(data, resolution):
    """
    Resolution series to reduce resolution uniformly.

    Parameters
    ----------
    data : Calliope model data
    resolution : int or float
        Resolution (in hours) to downsample all data to.

    """
    res_length = resolution / data.time_res_static
    df = data.r[list(data.r.keys())[0]]  # Grab length of data from any table
    summarize = pd.Series(-1, index=list(range(len(df))))
    # Set to 0 (keep timestep) for the given resolution
    for index, item in summarize.items():
        if index % res_length == 0:
            summarize.at[index] = resolution
    summarize.name = 'resolution_series'
    return summarize


def resolution_series_to_mask(resolution_series):
    """
    Turns a resolution series into a mask.

    """
    mask = resolution_series
    mask[mask != 0] = 1
    mask.name = 'mask'
    return mask



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

cli.py
~~~~~~

Command-line interface.

"""

import contextlib
import datetime
import logging
import os
import shutil
import sys
import traceback

import click

from . import core
from .parallel import Parallelizer


STRF = '%Y-%m-%d %H:%M:%S'

_debug = click.option('--debug', is_flag=True, default=False,
                      help='Print debug information when encountering errors.')
_pdb = click.option('--pdb', is_flag=True, default=False,
                    help='If used together with --debug, drop into interactive '
                         'debugger on encountering errors.')


@contextlib.contextmanager
def format_exceptions(debug=False, pdb=False, start_time=None):
    try:
        yield
    except Exception as e:
        if debug:
            traceback.print_exc()
            if pdb:
                import pdb
                pdb.post_mortem(e.__traceback__)
        else:
            stack = traceback.extract_tb(e.__traceback__)
            # Get last stack trace entry still in Calliope
            last = [i for i in stack if 'calliope' in i[0]][-1]
            err_string = 'Error in {}, {}:{}'.format(last[2], last[0], last[1])
            click.secho(err_string, fg='red')
            click.secho('\n' + str(e))
            if start_time:
                print_end_time(start_time, msg='aborted due to an error')
        sys.exit(1)


def print_end_time(start_time, msg='complete'):
    end_time = datetime.datetime.now()
    secs = round((end_time - start_time).total_seconds(), 1)
    tend = end_time.strftime(STRF)
    print('\nCalliope run {}. '
          'Elapsed: {} seconds (time at exit: {})'.format(msg, secs, tend))


@click.group()
def cli():
    """Calliope: a multi-scale energy systems (MUSES) modeling framework"""
    pass


@cli.command(short_help='create model')
@click.argument('path')
def new(path):
    """
    Create new model at the given PATH, based on the included example model.
    The directory must not yet exist, and intermediate directories will
    be created automatically.
    """
    # Copies the included example model
    example_model = os.path.join(os.path.dirname(__file__), 'example_model')
    click.echo('Creating new model in: {}'.format(path))
    shutil.copytree(example_model, path)


@cli.command(short_help='directly run single model')
@click.argument('run_config')
@_debug
@_pdb
def run(run_config, debug, pdb):
    """Execute the given RUN_CONFIG run configuration file."""
    logging.captureWarnings(True)
    start_time = datetime.datetime.now()
    with format_exceptions(debug, pdb, start_time):
        tstart = start_time.strftime(STRF)
        print('Calliope run starting at {}\n'.format(tstart))
        model = core.Model(config_run=run_config)
        model.time_format = STRF
        model.verbose = True  # Enables some print calls inside Model
        model_name = model.config_model.get_key('name', default='None')
        run_name = model.config_run.get_key('name', default='None')
        print('Model name:   {}'.format(model_name))
        print('Run name:     {}'.format(run_name))
        num_techs = (len(model.config_model.techs)
                     - len(core.get_default_techs()))
        msize = '{x} locations, {y} technologies, {t} timesteps'.format(
            x=len(model.config_model.locations),
            y=num_techs,
            t=len(model.data._dt))
        print('Model size:   {}'.format(msize))
        model.config_run.set_key('output.save', True)  # Always save output
        model.run()
        print_end_time(start_time)


@cli.command(short_help='generate parallel runs')
@click.argument('run_config')
@click.argument('path', default='runs')
@click.option('--silent', is_flag=True, default=False,
              help='Be less verbose.')
@_debug
@_pdb
def generate(run_config, path, silent, debug, pdb):
    """
    Generate parallel runs based on the given RUN_CONFIG configuration
    file, saving them in the given PATH, which is a path to a
    directory that must not yet exist (PATH defaults to 'runs'
    if not specified).
    """
    logging.captureWarnings(True)
    with format_exceptions(debug, pdb):
        parallelizer = Parallelizer(target_dir=path, config_run=run_config)
        if not silent and 'name' not in parallelizer.config.parallel:
            click.echo('`' + run_config +
                       '` does not specify a `parallel.name`' +
                       'and was skipped.')
            return
        click.echo('Generating runs from config '
                   '`{}` inside `{}`'.format(run_config, path))
        parallelizer.generate_runs()



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

read.py
~~~~~~~

Functions to read saved model results.

"""

import glob
import logging
import os

import pandas as pd

from .utils import AttrDict


REQUIRED_KEYS = ['capacity_factor', 'costs', 'levelized_cost',
                 'metadata', 'node', 'parameters',
                 'shares', 'summary', 'time_res', 'totals', 'config_model',
                 'config_run']  # 'data/...' keys are not required


def read_hdf(hdf_file, tables_to_read=None):
    """Read model solution from HDF file"""
    store = pd.HDFStore(hdf_file, mode='r')
    solution = AttrDict()
    if not tables_to_read:
        # Make sure leading/trailing '/' are removed from keys,
        # and don't read the 'config' key (yet)
        tables_to_read = [k.strip('/') for k in store.keys()
                          if k != 'config']
    for k in tables_to_read:
        solution[k] = store.get(k)
    # Also add model and run config to the solution object, which are stored
    # as strings in a Series in the 'config' key
    for k in ['config_model', 'config_run']:
        try:
            solution[k] = AttrDict.from_yaml_string(store.get('config')[k])
        except KeyError:
            pass  # Ignore missing 'config' key here, we log missing below
    # Check if any keys are missing
    missing_keys = set(REQUIRED_KEYS) - set(solution.keys())
    if len(missing_keys) > 0:
        logging.warning('HDF file {} missing keys: '
                        '{}'.format(hdf_file, missing_keys))
    store.close()
    return solution


def read_csv(directory, tables_to_read=None):
    solution = AttrDict()
    if not tables_to_read:
        tables_to_read = glob.glob(directory + '/*.csv')
        if len(tables_to_read) == 0:
            raise IOError('No CSV files found')
        # Only keep basenames without extension
        tables_to_read = [os.path.splitext(os.path.basename(f))[0]
                          for f in tables_to_read]
    for f in tables_to_read:
        src = os.path.join(directory, f + '.csv')
        df = pd.read_csv(src, index_col=0, parse_dates=True)
        # If 'minor' is in columns, we have a flattened panel!
        if 'minor' in df.columns:
            df['major'] = df.index
            df = df.set_index(['major', 'minor']).to_panel()
        solution[f] = df
    return solution


def _detect_format(directory):
    """Autodetects format, falling back to CSV if it can't find HDF"""
    if os.path.exists(os.path.join(directory, 'solution.hdf')):
        return 'hdf'
    else:
        return 'csv'


def read_dir(directory, tables_to_read=None):
    """Combines output files from `directory` and return an AttrDict
    containing them all.

    If a solution is missing or there is an error reading it, an empty
    AttrDict is added to the results in its stead and the error is logged.

    """
    results = AttrDict()
    results.iterations = pd.read_csv(os.path.join(directory, 'iterations.csv'),
                                     index_col=0)
    results.solutions = AttrDict()
    for i in results.iterations.index.tolist():
        iteration_dir = os.path.join(directory, '{:0>4d}'.format(i))
        fmt = _detect_format(iteration_dir)
        logging.debug('Iteration: {}, Format detected: {}'.format(i, fmt))
        try:
            if fmt == 'hdf':
                hdf_file = os.path.join(iteration_dir, 'solution.hdf')
                logging.debug('Read HDF: {}'.format(hdf_file))
                results.solutions[i] = read_hdf(hdf_file, tables_to_read)
            else:
                results.solutions[i] = read_csv(iteration_dir, tables_to_read)
                logging.debug('Read CSV: {}'.format(iteration_dir))
        except IOError as err:
            logging.warning('I/O error in `{}` at iteration `{}`'
                            ': {}'.format(iteration_dir, i, err))
            results.solutions[i] = AttrDict()  # add an empty entry
            continue
    return results



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

ramping.py
~~~~~~~~~~

Ramping constraints.

"""

import pyomo.core as po


def ramping_rate(model):
    """Depends on: node_energy_balance, node_constraints_build"""
    m = model.m
    time_res = model.data.time_res_series

    # Constraint rules
    def _ramping_rule(m, y, x, t, direction):
        # e_ramping: Ramping rate [fraction of installed capacity per hour]
        ramping_rate = model.get_option(y + '.constraints.e_ramping')
        if ramping_rate is False:
            # If the technology defines no `e_ramping`, we don't build a
            # ramping constraint for it!
            return po.Constraint.NoConstraint
        else:
            # No constraint for first timestep
            # NB: From Pyomo 3.5 to 3.6, order_dict became zero-indexed
            if m.t.order_dict[t] == 0:
                return po.Constraint.NoConstraint
            else:
                carrier = model.get_option(y + '.carrier')
                diff = ((m.es_prod[carrier, y, x, t]
                         + m.es_con[carrier, y, x, t]) / time_res.at[t]
                        - (m.es_prod[carrier, y, x, model.prev(t)]
                           + m.es_con[carrier, y, x, model.prev(t)])
                        / time_res.at[model.prev(t)])
                max_ramping_rate = ramping_rate * m.e_cap[y, x]
                if direction == 'up':
                    return diff <= max_ramping_rate
                else:
                    return -1 * max_ramping_rate <= diff

    def c_ramping_up_rule(m, y, x, t):
        return _ramping_rule(m, y, x, t, direction='up')

    def c_ramping_down_rule(m, y, x, t):
        return _ramping_rule(m, y, x, t, direction='down')

    # Constraints
    m.c_ramping_up = po.Constraint(m.y, m.x, m.t, rule=c_ramping_up_rule)
    m.c_ramping_down = po.Constraint(m.y, m.x, m.t, rule=c_ramping_down_rule)



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

analysis_utils.py
~~~~~~~~~~~~~~~~~

Helper functions for analysis.py.

"""

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass  # This is logged in analysis.py
import numpy as np
import pandas as pd


def legend_on_right(ax, style='default', artists=None, labels=None, **kwargs):
    """Draw a legend on outside on the right of the figure given by ``ax``"""
    box = ax.get_position()
    # originally box.width * 0.8 but 1.0 solves some problems
    # it just means that the box becomes wider, which is ok though!
    ax.set_position([box.x0, box.y0, box.width * 1.0, box.height])
    if style == 'square':
        artists, labels = get_square_legend(ax.legend())
        l = ax.legend(artists, labels, loc='center left',
                      bbox_to_anchor=(1, 0.5), **kwargs)
    elif style == 'custom':
        l = ax.legend(artists, labels, loc='center left',
                      bbox_to_anchor=(1, 0.5), **kwargs)
    else:
        l = ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), **kwargs)
    return l


def legend_below(ax, style='default', columns=5, artists=None, labels=None,
                 **kwargs):
    """Draw a legend on outside below the figure given by ``ax``"""
    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1,
                     box.width, box.height * 0.9])
    if style == 'square':
        artists, labels = get_square_legend(ax.legend())
        l = ax.legend(artists, labels, loc='upper center',
                      bbox_to_anchor=(0.5, -0.05), ncol=columns, **kwargs)
    elif style == 'custom':
        l = ax.legend(artists, labels, loc='upper center',
                      bbox_to_anchor=(0.5, -0.05), ncol=columns, **kwargs)
    else:
        l = ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
                      ncol=columns, **kwargs)
    return l


def get_square_legend(lgd):
    rects = [plt.Rectangle((0, 0), 1, 1,
             fc=l.get_color()) for l in lgd.get_lines()]
    labels = [l.get_label() for l in lgd.get_lines()]
    return (rects, labels)


def stack_plot(df, stack=None, figsize=None, colormap='jet', legend='default',
               ticks='daily', names=None, ax=None, leg_title=None,
               leg_fontsize=None, **kwargs):
    """
    if stack is None, the columns of the passed df are used

    legend can be 'default' or 'right', can set legend title with `leg_title`
    ticks can be 'hourly', 'daily', 'monthly'

    kwargs get passed to ax.stackplot()

    """
    if not stack:
        stack = df.columns
    if not ax:
        if not figsize:
            figsize = (16, 4)
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
    colors = plt.get_cmap(colormap)(np.linspace(0, 1.0, len(stack)))
    fills = ax.stackplot(df.index, df[stack].T, label=stack, colors=colors,
                         **kwargs)
    # Rename the tech stack with friendly names, if given, for legend plotting
    if names:
        stack = names
    # Legend via proxy artists
    # Based on https://github.com/matplotlib/matplotlib/issues/1943
    proxies = [plt.Rectangle((0, 0), 1, 1, fc=i.get_facecolor()[0])
               for i in fills]
    if legend == 'default':
        l = ax.legend(reversed(proxies), reversed(stack), title=leg_title)
    elif legend == 'right':
        l = legend_on_right(ax, artists=reversed(proxies),
                            labels=reversed(stack),
                            style='custom', title=leg_title)
    if leg_title and leg_fontsize:
        plt.setp(l.get_title(), fontsize=leg_fontsize)
    # Format x datetime axis
    # Based on http://stackoverflow.com/a/9627970/397746
    import matplotlib.dates as mdates
    if ticks == 'monthly':
        formatter = mdates.DateFormatter('%b %Y')
        locator = mdates.MonthLocator()
    if ticks == 'daily':
        formatter = mdates.DateFormatter('%d-%m-%Y')
        locator = mdates.DayLocator()
    if ticks == 'hourly':
        formatter = mdates.DateFormatter('%H:%M\n%d-%m-%Y')
        locator = mdates.HourLocator(byhour=[0])
        minor_formatter = mdates.DateFormatter('%H:%M')
        minor_locator = mdates.HourLocator(byhour=list(range(1, 24)))
        plt.gca().xaxis.set_minor_formatter(minor_formatter)
        plt.gca().xaxis.set_minor_locator(minor_locator)
    plt.gca().xaxis.set_major_formatter(formatter)
    plt.gca().xaxis.set_major_locator(locator)
    return ax


def _get_query_string(types):
    formatted_types = ['type == "{}"'.format(t) for t in types]
    query_string = ' | '.join(formatted_types)
    return query_string


def _get_ranges(dates):
    # Modified from http://stackoverflow.com/a/6934267/397746
    while dates:
        end = 1
        timedelta = dates[end] - dates[end - 1]
        try:
            while dates[end] - dates[end - 1] == timedelta:
                end += 1
        except IndexError:
            pass

        yield (dates[0], dates[end - 1])
        dates = dates[end:]


def plot_graph_on_map(config_model, G=None,
                      edge_colors=None, edge_labels=None,
                      figsize=(15, 15), fontsize=9,
                      arrow_style='->',
                      rotate_labels=False,
                      bounds=None,
                      scale_left_distance=0.05,
                      scale_bottom_distance=0.05):
    from mpl_toolkits.basemap import Basemap
    import networkx as nx
    from calliope.lib import nx_pylab

    # Set up basemap
    if not bounds:
        bounds = config_model.metadata.map_boundary
    bounds_width = bounds[2] - bounds[0]  # lon --> width
    bounds_height = bounds[3] - bounds[1]  # lat --> height
    m = Basemap(projection='merc', ellps='WGS84',
                llcrnrlon=bounds[0], llcrnrlat=bounds[1],
                urcrnrlon=bounds[2], urcrnrlat=bounds[3],
                lat_ts=bounds[1] + bounds_width / 2,
                resolution='i',
                suppress_ticks=True)

    # Node positions
    pos = config_model.metadata.location_coordinates
    pos = {i: m(pos[i][1], pos[i][0]) for i in pos}  # Flip lat, lon to x, y!

    # Create plot
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111, axisbg='w', frame_on=False)
    m.drawmapboundary(fill_color=None, linewidth=0)
    m.drawcoastlines(linewidth=0.2, color='#626262')

    # Draw the graph
    if G:
        # Using nx_pylab to be able to set zorder below the edges
        nx_pylab.draw_networkx_nodes(G, pos, node_color='#CCCCCC',
                                     node_size=300, zorder=0)

        # Using nx_pylab from lib to get arrow_style option
        nx_pylab.draw_networkx_edges(G, pos, width=3,
                                     edge_color=edge_colors,
                                     # This works for edge_use
                                     edge_vmin=0.0, edge_vmax=1.0,
                                     edge_cmap=plt.get_cmap('seismic'),
                                     arrows=True, arrow_style=arrow_style)

        # bbox = dict(color='white', alpha=0.5, edgecolor=None)
        labels = nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels,
                                              rotate=rotate_labels,
                                              font_size=fontsize)

    # Add a map scale
    scale = m.drawmapscale(
        bounds[0] + bounds_width * scale_left_distance,
        bounds[1] + bounds_height * scale_bottom_distance,
        bounds[0], bounds[1],
        100,
        barstyle='simple', labelstyle='simple',
        fillcolor1='w', fillcolor2='#555555',
        fontcolor='#555555', fontsize=fontsize
    )

    return ax, m



import pytest
import tempfile

from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver, _add_test_path


def create_and_run_model(override, iterative_warmstart=False):
    locations = """
        locations:
            1:
                level: 0
                within:
                techs: ['ccgt', 'demand_electricity']
                override:
                    ccgt:
                        constraints:
                            e_cap_max: 100
                    demand_electricity:
                        x_map: '1: demand'
                        constraints:
                            r: file=demand-blocky_r.csv
        links:
    """
    config_run = """
        mode: plan
        model: [{techs}, {locations}]
    """
    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    with tempfile.NamedTemporaryFile() as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override,
                                    path=_add_test_path('common/t_time'))
    model.run(iterative_warmstart)
    return model


class TestModel:
    def test_model_subset_t_from_middata(self):
        override = """
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        df = model.solution.node
        assert df.loc['e:power', 'ccgt', :, :].sum().sum() == 720

    def test_model_time_res_uniform(self):
        override = """
            time:
                resolution: 24
        """
        model = create_and_run_model(override)
        assert len(model.data.time_res_series) == 4
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        df = model.solution.node
        assert df.loc['e:power', 'ccgt', :, :].sum().sum() == 1320

    def test_model_time_res_uniform_subset_t_from_start(self):
        override = """
            time:
                resolution: 24
            subset_t: ['2005-01-01', '2005-01-02']
        """
        model = create_and_run_model(override)
        assert len(model.data.time_res_series) == 2
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        df = model.solution.node
        assert df.loc['e:power', 'ccgt', :, :].sum().sum() == 600

    def test_model_time_res_uniform_subset_t_from_middata(self):
        override = """
            time:
                resolution: 24
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert len(model.data.time_res_series) == 2
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        df = model.solution.node
        assert df.loc['e:power', 'ccgt', :, :].sum().sum() == 720

    def test_model_subset_t_operational(self):
        override = """
            mode: operate
            override:
                opmode:
                    horizon: 24  # Optimization period length (hours)
                    window: 12  # Operation period length (hours)
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        df = model.solution.node
        assert df.loc['e:power', 'ccgt', :, :].sum().sum() == 720

    def test_model_time_res_uniform_subset_t_operational(self):
        override = """
            mode: operate
            override:
                opmode:
                    horizon: 48  # Optimization period length (hours)
                    window: 24  # Operation period length (hours)
            time:
                resolution: 12
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert len(model.data.time_res_series) == 4
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        df = model.solution.node
        assert df.loc['e:power', 'ccgt', :, :].sum().sum() == 720



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

time_masks.py
~~~~~~~~~~~~~

Defines time masks to be used with TimeSummarizer (see time_tools.py).

A mask function returns a pandas Series with the same index as the input
data that is 0 where the original resolution should be retained (unmasked
areas) and 1 where it should be reduced (masked areas).

The name of a mask must always be 'mask'.

"""

import pandas as pd


def mask_relative_difference(data, target1, target2, days=5,
                             padding_days=1, per_zone=False):
    """
    Masks everywhere except the ``days`` number of days where the
    resource of the two technologies ``target1`` and ``target2`` is most
    different.

    Parameters
    ----------
    data : Calliope model data
    target1 : str
    target2 : str
    days : int, default 5
    padding_days : int, default 1
    per_zone: bool, default False

    """
    data_length = len(data._dt)

    def _get_most_different(data, target1, target2,
                            days, padding_days, target_zone=None):
        # Get length of a day in timesteps
        day_len = int(24 / data.time_res_data)
        targets = (target1, target2)
        relative_series = []
        for t in targets:
            if not target_zone:
                t_series = data.r[t].sum(1)
            else:
                t_series = data.r[t].loc[:, target_zone]
            if t_series.max() < 0:  # Assume this is a demand series and <0
                relative = t_series / t_series.min()
            else:
                relative = t_series / t_series.max()
            relative_series.append(relative)

        diff = (relative_series[0] - relative_series[1]).abs()

        most_different = diff.sort(ascending=False, inplace=False)
        datetimes = data._dt.loc[most_different.index]
        # `result` are the datetimes for the found days (indexed by timestep index)
        result = datetimes.dt.date.drop_duplicates().head(days)

        periods = []
        for index, r in result.iteritems():
            # `selected` is the index of the first timestamp in the chosen day
            selected = data._dt[data._dt.dt.date == r].index[0]
            start = selected - padding_days * day_len
            end = selected + padding_days * day_len
            periods.append((start, end))
        return periods

    if not per_zone:
        periods = _get_most_different(data, target1, target2, days,
                                      padding_days)
    else:
        periods = []
        for zone in data._x:
            p = _get_most_different(data, target1, target2, days,
                                    padding_days, target_zone=zone)
            periods.extend(p)

    # Mask where everything is 1 (reduce resolution) by default
    mask = pd.Series(1, index=list(range(data_length)))
    # For the desired time periods, change the mask to native resolution (0)
    for p in periods:
        mask.loc[p[0]:p[1] - 1] = 0
    mask.name = 'mask'
    return mask


def mask_zero(data, tech, var='r', locations=None):
    """
    Mask where ``var`` for the technology ``tech``
    across the given list of ``locations`` is zero.

    ``var`` defaults to ``r``.

    If ``locations`` not given, uses all available locations.

    """
    df = data[var][tech]
    if locations:
        df = df.loc[:, locations]
    # Retain original resolution everywhere by default (0)
    mask = pd.Series(0, index=list(range(len(df))))
    # Sum over all locations to find those times where the data==0 everywhere
    mask[df.sum(1) <= 0] = 1
    mask.name = 'mask'
    return mask


def mask_extreme(data, tech, var='r', how='max',
                 length=24, locations=None, padding=None):
    """
    Mask everywhere except the ``length`` where ``var`` for the technology
    ``tech`` across the given list of ``locations`` is either minmal
    or maximal.

    Parameters
    ----------
    data : Calliope model data
    tech : str
        Technology whose `var` to find extreme for.
    var : str, default 'r'
    how : str, default 'max'
        'max' or 'min'
    length : int, default 24
        Timesteps to mask.
    locations : list, default None
        List of locations to use, if None, uses all available locations.
    padding : int, default None
        Pad beginning and end of the unmasked area by the number of
        timesteps given.

    """
    df = data[var][tech]
    if locations:
        df = df.loc[:, locations]
    df = df.sum(axis=1)
    totals = []
    for i in range(0, len(df), length):
        totals.append(df[i:(i + 1) * length].sum(axis=0))
    totals = pd.Series(totals)
    if how == 'max':
        total_i = totals.argmax()
    elif how == 'min':
        total_i = totals.argmin()
    mask = pd.Series(1, index=list(range(len(df))))
    ifrom = total_i * length
    ito = (total_i + 1) * length - 1
    if padding:
        ifrom = ifrom - padding
        ito = ito + padding
        # Ensure ifrom and ito remain within bounds
        if ifrom < 0:
            ifrom = 0
        if ito > len(df):
            ito = len(df)
    mask.loc[ifrom:ito] = 0
    mask.name = 'mask'
    return mask


def mask_extreme_week(data, tech, var='r', how='sum', what='min'):
    """
    Mask everywhere except the week containing the day where ``var`` of
    ``tech`` is minimal (what='min') or maximal (what='max') across the
    sum (how='sum') of or the mode (most) of locations (how='mode').

    Parameters
    ----------
    data : Calliope model data
    tech : str
        Technology whose `var` to find extreme week for.
    var : str, default 'r'
    how : str, default 'sum'
        'sum' or 'mode'
    what : str, default 'min'
        'min' or 'max'

    """
    df = data[var][tech]
    # Get length of a day in timesteps
    day_len = int(24 / data.time_res_data)
    # Get day-wise sums
    dff_index = list(range(0, len(df), day_len))
    dff = pd.rolling_sum(df, window=day_len).reindex(dff_index)
    # If what is 'min', this will get the 'idxmin' attribute (a method),
    # similar for 'max', else most likely raise an error!
    idx_extr = lambda x: getattr(x, 'idx{}'.format(what))
    if how == 'mode':
        # Find day where var/tech is min/max across the most locations
        selected = idx_extr(dff)().mode()[0]
    elif how == 'sum':
        # Find day where var/tech is min/max across all locations
        selected = idx_extr(dff.sum(axis=1))()
    d = data._dt.at[selected]
    # Determine the range for the calendar week
    # (7 days) to keep at full resolution
    week_start = selected - day_len * d.dayofweek
    week_end = selected + day_len * (7 - d.dayofweek)
    # Mask where everything is 1 (reduce resolution) by default
    mask = pd.Series(1, index=list(range(len(df))))
    # For the desired week, change the mask to native resolution (0)
    mask.loc[week_start:week_end - 1] = 0
    mask.name = 'mask'
    return mask



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

core.py
~~~~~~~

Core model functionality via the Model class.

"""

import datetime
import functools
import importlib
import inspect
import itertools
import logging
import os
import random
import shutil
import sys
import time
import warnings

import pyomo.opt as popt
import pyomo.core as po
import pyomo.environ  # Necessary for solver plugins etc.
import numpy as np
import pandas as pd
from pyutilib.services import TempfileManager

from . import exceptions
from . import constraints
from . import locations
from . import output
from . import transmission
from . import time_masks
from . import time_tools
from . import utils

# Enable simple format when printing ModelWarnings
formatwarning_orig = warnings.formatwarning


def _formatwarning(message, category, filename, lineno, line=None):
    """Formats ModelWarnings as "Warning: message" without extra crud"""
    if category == exceptions.ModelWarning:
        return 'Warning: ' + str(message) + '\n'
    else:
        return formatwarning_orig(message, category, filename, lineno, line)

warnings.formatwarning = _formatwarning


@functools.lru_cache(maxsize=1)
def get_default_techs(foo=0):
    """
    Get list of techs pre-defined in defaults.yaml.

    The foo=0 parameter makes sure that lru_cache has an argument to cache,
    the function must always be called as get_default_techs() with no
    arguments, ensuring that the values are only read from disk once and
    then cached.

    """
    module_config = os.path.join(os.path.dirname(__file__), 'config')
    o = utils.AttrDict.from_yaml(os.path.join(module_config, 'defaults.yaml'))
    return list(o.techs.keys())


def _load_function(source):
    """
    Returns a function from a module, given a source string of the form:

        'module.submodule.subsubmodule.function_name'

    """
    module_string, function_string = source.rsplit('.', 1)
    modules = [i for i in sys.modules.keys() if 'calliope' in i]
    # Check if module already loaded, if so, don't re-import it
    if (module_string in modules):
        module = sys.modules[module_string]
    elif ('calliope.' + module_string) in modules:
        module = sys.modules['calliope.' + module_string]
    # Else load the module
    else:
        try:
            module = importlib.import_module(module_string)
        except ImportError:
            module = importlib.import_module('calliope.' + module_string)
    return getattr(module, function_string)


def get_model_config(cr, config_run_path, adjust_data_path=None,
                     insert_defaults=True):
    """
    cr is the run configuration AttrDict,
    config_run_path the path to the run configuration file

    If ``adjust_data_path`` is given, the data_path setting is adjusted
    using the given path, else, it is forced to an absolute path.

    If ``insert_defaults`` is False, the default settings from
    defaults.yaml will not be included, which is necessary when
    generating model settings file for parallel runs.

    """
    # Ensure 'model' key is a list
    if not isinstance(cr.model, list):
        cr.model = [cr.model]

    # Interpret relative config paths as relative to run.yaml
    cr.model = [utils.relative_path(i, config_run_path) for i in cr.model]

    # Load defaults from module path
    module_conf = os.path.join(os.path.dirname(__file__), 'config')
    o = utils.AttrDict.from_yaml(os.path.join(module_conf, 'defaults.yaml'))

    # If defaults should not be inserted, replace the loaded AttrDict
    # with an empty one (a bit of a hack, but we also want the
    # default_techs list so we need to load the AttrDict anyway)
    if not insert_defaults:
        o = utils.AttrDict()
        o.techs = utils.AttrDict()

    # Load all additional files, continuously checking consistency
    for path in cr.model:
        new_o = utils.AttrDict.from_yaml(path)
        if 'techs' in list(new_o.keys()):
            overlap = set(get_default_techs()) & set(new_o.techs.keys())
            if overlap:
                e = exceptions.ModelError
                raise e('Trying to re-define a default technology in '
                        '{}: {}'.format(path, list(overlap)))
        # Interpret data_path as relative to `path`  (i.e the currently
        # open model config file), unless `adjust_data_path` is given
        if 'data_path' in new_o:
            if adjust_data_path:
                new_o.data_path = os.path.join(adjust_data_path,
                                               new_o.data_path)
            else:
                new_o.data_path = utils.relative_path(new_o.data_path, path)
        # The input files are allowed to override defaults
        o.union(new_o, allow_override=True)

    return o


class Model(object):
    """
    Calliope: a multi-scale energy systems (MUSES) modeling framework

    Parameters
    ----------
    config_run : str or AttrDict, default None
        Path to YAML file with run settings, or AttrDict containing run
        settings. If not given, the included default run and model
        settings are used.
    override : AttrDict, default None
        Provide any additional options or override options from
        ``config_run`` by passing an AttrDict of the form
        ``{'model_settings': 'foo.yaml'}``. Any option possible in
        ``run.yaml`` can be specified in the dict, inluding ``override.``
        options.
    initialize_config_only : bool, default False
        If True, the model will not be runnable, and only the basic
        model configuration will be initialized, but time series data
        will be read or initialized, which can be useful for faster
        debugging or processing the configuration.

    """
    def __init__(self, config_run=None, override=None,
                 initialize_config_only=False):
        super(Model, self).__init__()
        self.verbose = False
        self.time_format = '%Y-%m-%d %H:%M:%S'
        self.debug = utils.AttrDict()
        self.initialize_configuration(config_run, override)
        # Other initialization tasks
        self.data = utils.AttrDict()
        # Initialize the option getter
        self._get_option = utils.option_getter(self.config_model, self.data)
        self.initialize_parents()
        self.initialize_sets()
        if not initialize_config_only:
            self.read_data()
            self.mode = self.config_run.mode
            self.initialize_availability()
            self.initialize_time()

    def override_model_config(self, override_dict):
        o = self.config_model
        od = override_dict
        if 'data_path' in od.keys_nested():
            # If run_config overrides data_path, interpret it as
            # relative to the run_config file's path
            od['data_path'] = utils.relative_path(od['data_path'],
                                                  self.config_run_path)
        o.union(od, allow_override=True, allow_replacement=True)

    def initialize_configuration(self, config_run, override):
        self.flush_option_cache()
        # Load run configuration
        if not config_run:
            config_run = os.path.join(os.path.dirname(__file__),
                                      'example_model', 'run.yaml')
        if isinstance(config_run, str):
            # 1) config_run is a string, assume it's a path
            cr = utils.AttrDict.from_yaml(config_run)
            # self.run_id is used to set an output folder for logs, if
            # debug.keep_temp_files is set to True
            self.run_id = os.path.splitext(os.path.basename(config_run))[0]
            self.config_run_path = config_run
        else:
            # 2) config_run is not a string, assume it's an AttrDict
            cr = config_run
            assert isinstance(cr, utils.AttrDict)
            # we have no filename so we just use current date/time
            self.run_id = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            # Use current working directory as config_run path
            self.config_run_path = os.getcwd()
        self.config_run = cr
        if override:
            assert isinstance(override, utils.AttrDict)
            cr.union(override, allow_override=True, allow_replacement=True)
        # If manually specify a run_id in debug, overwrite the generated one
        if 'debug.run_id' in cr.keys_nested():
            self.run_id = cr.debug.run_id
        self.config_model = get_model_config(cr, self.config_run_path)
        # Override config_model settings if specified in config_run
        # 1) Via 'model_override', which is the path to a YAML file
        if 'model_override' in cr:
            override_path = utils.relative_path(cr.model_override,
                                                self.config_run_path)
            override_dict = utils.AttrDict.from_yaml(override_path)
            self.override_model_config(override_dict)
        # 2) Via 'override', which is an AttrDict
        if ('override' in cr
                and isinstance(cr.override, utils.AttrDict)):
            self.override_model_config(cr.override)
        # Initialize locations
        locs = self.config_model.locations
        self.config_model.locations = locations.process_locations(locs)
        # As a final step, flush the option cache
        self.flush_option_cache()

    def get_timeres(self, verify=False):
        """Returns resolution of data in hours. Needs a properly
        formatted ``set_t.csv`` file to work.

        If ``verify=True``, verifies that the entire file is at the same
        resolution. ``model.get_timeres(verify=True)`` can be called
        after Model initialization to verify this.

        """
        path = self.config_model.data_path
        df = pd.read_csv(os.path.join(path, 'set_t.csv'), index_col=0,
                         header=None, parse_dates=[1])
        seconds = (df.iat[0, 0] - df.iat[1, 0]).total_seconds()
        if verify:
            for i in range(len(df) - 1):
                assert ((df.iat[i, 0] - df.iat[i+1, 0]).total_seconds()
                        == seconds)
        hours = abs(seconds) / 3600
        return hours

    def initialize_availability(self):
        # Availability in time/space is only used if dynamic timestep
        # adjustment is invoked
        d = self.data
        d['a'] = utils.AttrDict()
        # Append 'a' to d.params, so that it is included in parameter updates!
        d.params.append('a')
        # Fill in default values for a, so that something is there even in
        # case no dynamic timestepper is called
        for y in d._y_def_r:
            avail = self.get_option(y + '.constraints.availability')
            d.a[y] = pd.DataFrame(avail, index=d._dt.index, columns=d._x)

    def initialize_time(self):
        """
        Performs time resolution reduction, if set up in configuration

        """
        cr = self.config_run
        s = time_tools.TimeSummarizer()
        time = cr.get_key('time', default=False)
        if time:
            if ('masks' in time or 'resolution' in time) and 'file' in time:
                e = exceptions.ModelError
                raise e('`time.masks` or `time.resolution` and `time.file` '
                        'cannot be given at the same time.')
            if 'masks' in time:
                masks = []
                # time.masks is a list of {'function': .., 'options': ..} dicts
                for entry in time.masks:
                    entry = utils.AttrDict(entry)
                    func_string = 'time_masks.' + entry.function
                    mask_func = _load_function(func_string)
                    mask_opts = entry.get_key('options', default=False)
                    if mask_opts:
                        mask = mask_func(self.data, **mask_opts)
                    else:
                        mask = mask_func(self.data)
                    masks.append(mask)
                if 'resolution' in time:
                    mask_res = time.resolution
                else:
                    mask_res = None
                converter = time_tools.masks_to_resolution_series
                series = converter(masks, how='or', resolution=mask_res)
            elif 'resolution' in time:
                if self.data.time_res_static == time.resolution:
                    # If time.resolution is set to the resolution that
                    # the data is already in, do nothing
                    return
                else:
                    # Otherwise, prepare an appropriate time resolution
                    # series for resampling the data
                    getter = time_tools.resolution_series_uniform
                    series = getter(self.data, time.resolution)
            elif 'file' in time:
                res_file = utils.relative_path(cr.time.file,
                                               self.config_run_path)
                series = pd.read_csv(res_file, index_col=0, header=None)[1] \
                           .astype(int)[self.slice]
            # Silently ignore if neither time.file or time.function given
            else:
                return
            s.dynamic_timestepper(self.data, series)

    def prev(self, t):
        """Using the timesteps set of this model instance, return `t-1`,
        even if the set is not continuous.

        E.g., if t is [0, 1, 2, 6, 7, 8], model.prev(6) will return 2.

        """
        # Create an index to look up t, and save it for later use
        try:
            # Check if _t_index exists
            self._t_index.name
        except AttributeError:
            self._t_index = pd.Index(self.data._dt.index)
        # Get the location of t in the index and use it to retrieve
        # the desired value, raising an error if it's <0
        loc = self._t_index.get_loc(t) - 1
        if loc >= 0:
            return self.data._dt.index[loc]
        else:
            e = exceptions.ModelError
            raise e('Attempted to get a timestep earlier than the first one.')

    def get_option(self, option, x=None, default=None,
                   ignore_inheritance=False):
        """
        Retrieves options from model settings for the given tech,
        falling back to the default if the option is not defined for the
        tech.

        If ``x`` is given, will attempt to use location-specific override
        from the location matrix first before falling back to model-wide
        settings.

        If ``default`` is given, it is used as a fallback if no default value
        can be found in the regular inheritance chain. If ``default`` is None
        and the regular inheritance chain defines no default, an error
        is raised.

        If ``ignore_inheritance`` is True, the default is immediately used
        instead of a search through the inheritance chain if the option
        has not been set for the given tech.

        If the first segment of the option contains ':', it will be
        interpreted as implicit tech subsetting: e.g. asking for
        'hvac:r1' implicitly uses 'hvac:r1' with the parent 'hvac', even
        if that has not been defined, to search the option inheritance
        chain.

        Examples:

        * ``model.get_option('ccgt.costs.om_var')``
        * ``model.get_option('csp.weight')``
        * ``model.get_option('csp.r', x='33')``
        * ``model.get_option('ccgt.costs.om_var',\
          default='defaults.costs.om_var')``

        """
        key = (option, x, default, ignore_inheritance)
        try:
            result = self.option_cache[key]
        except KeyError:
            # self._get_option is defined inside __init__
            result = self.option_cache[key] = self._get_option(*key)
        return result

    def set_option(self, option, value, x=None):
        """
        Set ``option`` to ``value``. Returns None on success.

        A default can be set by passing an option like
        ``defaults.constraints.e_eff``.

        """
        o = self.config_model
        d = self.data
        if x is None:
            o.set_key('techs.' + option, value)
        else:  # Setting a specific x
            d.locations.at[x, '_override.' + option] = value
        self.flush_option_cache()

    def flush_option_cache(self):
        self.option_cache = {}

    def get_name(self, y):
        try:
            return self.get_option(y + '.name')
        except exceptions.OptionNotSetError:
            return y

    def get_carrier(self, y):
        return self.get_option(y + '.carrier')

    def get_weight(self, y):
        return self.get_option(y + '.stack_weight')

    def get_color(self, y):
        color = self.get_option(y + '.color')
        if color is False:
            # If no color defined, choose one by seeding random generator
            # with the tech name to get pseudo-random one
            random.seed(y)
            r = lambda: random.randint(0, 255)
            color = '#{:0>2x}{:0>2x}{:0>2x}'.format(r(), r(), r())
        return color

    def get_source_carrier(self, y):
        source_carrier = self.get_option(y + '.source_carrier')
        if source_carrier:
            return source_carrier
        else:
            return None

    def get_parent(self, y):
        """Returns the abstract base technology from which ``y`` descends."""
        if y in self.data._y_transmission:
            return 'transmission'
        else:
            while True:
                parent = self.get_option(y + '.parent')
                if parent == 'defaults':
                    break
                y = parent
            return y

    def get_date_index(self, date):
        return self.data._dt[self.data._dt == date].index[0]

    def get_group_members(self, group, in_model=True, head_nodes_only=True,
                          expand_transmission=True):
        """
        Return the member technologies of a group. If ``in_model`` is
        True, only technologies (head nodes) in use in the current model
        are returned.

        Returns:
            * A list of group members if there are any.
            * If a group has no members (is only member of other
              groups, i.e. a head node), a list with a single item
              containing only the group/technology itself.
            * An empty list if the group is defined but not allowed
              in the current model.
            * None if the group doesn't exist.

        Other arguments:

            ``head_nodes_only`` : if True, don't return intermediate
                                  groups, i.e. technology definitions
                                  that are inherited from. Setting this
                                  to False only makes sense if in_model
                                  is also False, because in_model=True
                                  implies that only head nodes are
                                  returned.

            ``expand_transmission`` : if True, return in-model
                                      transmission technologies in the
                                      form ``tech:location``.

        """
        def _get(self, group, memberset):
            members = [i for i in self.parents if self.parents[i] == group]
            if members:
                for i, member in enumerate(members):
                    if not head_nodes_only:
                        memberset.add(member)
                    members[i] = _get(self, member, memberset)
                return members
            else:
                memberset.add(group)
        if not group in self.parents:
            return None
        memberset = set()
        _get(self, group, memberset)  # Fills memberset
        if in_model:
            memberset = set([y for y in memberset
                             if (y in self.data._y
                                 or y in self.data.transmission_y)])
            # Expand transmission techs
            if expand_transmission:
                for y in list(memberset):
                    if y in self.data.transmission_y:
                        memberset.remove(y)
                        memberset.update([yt
                                          for yt in self.data._y_transmission
                                          if yt.startswith(y + ':')])
        return list(memberset)

    @utils.memoize_instancemethod
    def get_eff_ref(self, var, y, x=None):
        """Get reference efficiency, falling back to efficiency if no
        reference efficiency has been set."""
        base = y + '.constraints.' + var
        try:
            eff_ref = self.get_option(base + '_eff_ref', x=x)
        except exceptions.OptionNotSetError:
            eff_ref = False
        if eff_ref is False:
            eff_ref = self.get_option(base + '_eff', x=x)
        # NOTE: Will cause errors in the case where (1) eff_ref is not defined
        # and (2) eff is set to "file". That is ok however because in this edge
        # case eff_ref should be manually set as there is no straightforward
        # way to derive it from the time series file.
        return eff_ref

    def scale_to_peak(self, df, peak, scale_time_res=True):
        """Returns the given dataframe scaled to the given peak value.

        If ``scale_time_res`` is True, the peak is multiplied by the model's
        time resolution. Set it to False to scale things like efficiencies.

        """
        # Normalize to (0, 1), then multiply by the desired maximum,
        # scaling this peak according to time_res
        if scale_time_res:
            adjustment = self.get_timeres()
        else:
            adjustment = 1
        if peak < 0:
            scale = float(df.min())
        else:
            scale = float(df.max())
        return (df / scale) * peak * adjustment

    def initialize_parents(self):
        o = self.config_model
        try:
            self.parents = {i: o.techs[i].parent for i in o.techs.keys()
                            if i != 'defaults'}
        except KeyError:
            tech = inspect.trace()[-1][0].f_locals['i']
            if 'parent' not in list(o.techs[tech].keys()):
                e = exceptions.ModelError
                raise e('Technology `' + tech + '` defines no parent!')
        # Verify that no technologies apart from the default technologies
        # inherit from 'defaults'
        for k, v in self.parents.items():
            if k not in get_default_techs() and v == 'defaults':
                e = exceptions.ModelError
                raise e('Tech `' + k + '` inherits from `defaults` but ' +
                        'should inherit from a built-in default technology.')
        # Verify that all parents are themselves actually defined
        for k, v in self.parents.items():
            if v not in list(o.techs.keys()):
                e = exceptions.ModelError
                raise e('Parent `' + v + '` of technology `' +
                        k + '` is not defined.')

    @utils.memoize_instancemethod
    def ischild(self, y, of):
        """Returns True if ``y`` is a child of ``of``, else False"""
        result = False
        while (result is False) and (y != 'defaults'):
            parent = self.parents[y]
            if parent == of:
                result = True
            y = parent
        return result

    def _initialize_transmission(self):
        o = self.config_model
        d = self.data
        # Add transmission technologies to y, if any defined
        if d._y_transmission:
            d._y.extend(d._y_transmission)
            # Add transmission tech columns to locations matrix
            for y in d._y_transmission:
                d.locations[y] = 0
            # Create representation of location-tech links
            tree = transmission.explode_transmission_tree(o.links, d._x)
            # Populate locations matrix with allowed techs and overrides
            if tree:
                for x in tree:
                    for y in tree[x]:
                        # Allow the tech
                        d.locations.at[x, y] = 1
                        # Add constraints if needed
                        for c in tree[x][y].keys_nested():
                            colname = '_override.' + y + '.' + c
                            if not colname in d.locations.columns:
                                d.locations[colname] = np.nan
                            d.locations.at[x, colname] = tree[x][y].get_key(c)

    def initialize_sets(self):
        o = self.config_model
        d = self.data
        path = o.data_path
        #
        # t: Timesteps set
        #
        table_t = pd.read_csv(os.path.join(path, 'set_t.csv'), header=None,
                              index_col=1, parse_dates=[1])
        table_t.columns = ['t_int']
        if self.config_run.get_key('subset_t', default=False):
            table_t = table_t.loc[self.config_run.subset_t[0]:
                                  self.config_run.subset_t[1]]
            self.slice = slice(table_t.iat[0, 0], table_t.iat[-1, 0] + 1)
        else:
            self.slice = slice(None)
        _t = pd.Series([int(t) for t in table_t['t_int'].tolist()])
        d._dt = pd.Series(table_t.index, index=_t.tolist())
        # First set time_res_data and time_res_static across all data
        # `time_res_data` never changes, so always reflects the spacing
        # of time step indices
        d.time_res_data = self.get_timeres()
        # `time_res_static` is updated after time resolution adjustments,
        # so does not reflect the spacing of time step indicex
        d.time_res_static = d.time_res_data
        d.time_res_native = 1  # In the beginning, time_res is native
        # From time_res_data, initialize time_res_series
        d.time_res_series = pd.Series(d.time_res_data, index=d._dt.index)
        # Last index t for which model may still use startup exceptions
        d.startup_time_bounds = d._dt.index[int(o.startup_time
                                                / d.time_res_data)]
        #
        # y: Technologies set
        #
        d._y = set()
        try:
            for k, v in o.locations.items():
                for y in v.techs:
                    if y in o.techs:
                        d._y.add(y)
                    else:
                        e = exceptions.ModelError
                        raise e('Location `{}` '
                                'uses undefined tech `{}`.'.format(k, y))
        except KeyError:
            e = exceptions.ModelError
            raise e('The region `' + k + '` does not allow '
                    'any technologies via `techs`. Must give '
                    'at least one technology per region.')
        d._y = list(d._y)
        if self.config_run.get_key('subset_y', default=False):
            d._y = [y for y in d._y if y in self.config_run.subset_y]
        # Subset of transmission technologies, if any defined
        # Used to initialized transmission techs further below
        # (not yet added to d._y here)
        if ('links' in o) and (o.links is not None):
            d._y_transmission = transmission.get_transmission_techs(o.links)
            d.transmission_y = list(set([list(v.keys())[0]
                                    for k, v in o.links.items()]))
        else:
            d._y_transmission = []
            d.transmission_y = []
        # Subset of conversion technologies
        d._y_conversion = [y for y in d._y if self.ischild(y, of='conversion')]
        # Subset of supply, demand, storage technologies
        d._y_pc = [y for y in d._y
                   if not self.ischild(y, of='conversion')
                   or self.ischild(y, of='transmission')]
        # Subset of technologies that define es_prod/es_con
        d._y_prod = ([y for y in d._y if not self.ischild(y, of='demand')]
                     + d._y_transmission)
        d._y_con = ([y for y in d._y if not self.ischild(y, of='supply')]
                    + d._y_transmission)
        # Subset of technologies that allow rb
        d._y_rb = [y for y in d._y
                   if self.get_option(y + '.constraints.allow_rb') is True]
        # Subset of technologies with parasitics (carrier efficiency != 1.0)
        d._y_p = [y for y in d._y
                  if self.get_option(y + '.constraints.c_eff') != 1.0]
        #
        # c: Carriers set
        #
        d._c = set()
        for y in d._y:  # Only add carriers for allowed technologies
            d._c.update([self.get_option(y + '.carrier')])
            if self.get_option(y + '.source_carrier'):
                d._c.update([self.get_option(y + '.source_carrier')])
        d._c = list(d._c)
        #
        # x: Locations set
        #
        d._x = list(o.locations.keys())
        if self.config_run.get_key('subset_x', default=False):
            d._x = [x for x in d._x if x in self.config_run.subset_x]
        #
        # Locations settings matrix and transmission technologies
        #
        d.locations = locations.generate_location_matrix(o.locations,
                                                         techs=d._y)
        # For simplicity, only keep the locations that are actually in set `x`
        d.locations = d.locations.loc[d._x, :]
        #
        # Initialize transmission technologies
        #
        self._initialize_transmission()
        #
        # self.data._y is now complete, ensure that all techs conform to the
        # rule that only "head" techs can be used in the model
        #
        for y in self.data._y:
            if self.get_option(y + '.parent') in self.data._y:
                e = exceptions.ModelError
                raise e('Only technologies without children can be used '
                        'in the model definition '
                        '({}, {}).'.format(y, self.get_option(y + '.parent')))
        #
        # k: Cost classes set
        #
        classes = [list(o.techs[k].costs.keys()) for k in o.techs
                   if k != 'defaults'  # Prevent 'default' from entering set
                   if 'costs' in o.techs[k]]
        # Flatten list and make sure 'monetary' is in it
        classes = ([i for i in itertools.chain.from_iterable(classes)]
                   + ['monetary'])
        d._k = list(set(classes))  # Remove any duplicates with a set roundtrip

    def read_data(self):
        """
        Read parameter data from CSV files, if needed. Data that may be
        defined in CSV files is read before generate_model() so that it
        only has to be read from disk once, even if generate_model() is
        repeatedly called.

        Note on indexing: if subset_t is set, a subset of the data is
        selected, and all data including their indices are subsetted.
        d._dt.index maps a simple [0, data length] index to the actual
        t index used.

        Data is stored in the `self.data`  for each
        `param` and technology `y`: ``self.data[param][y]``

        """
        @utils.memoize
        def _get_option_from_csv(filename):
            """Read CSV time series"""
            d_path = os.path.join(self.config_model.data_path, filename)
            # [self.slice] to do time subset if needed
            # Results in e.g. d.r_eff['csp'] being a dataframe
            # of efficiencies for each time step t at location x
            df = pd.read_csv(d_path, index_col=0)[self.slice]
            # Fill columns that weren't defined with NaN
            # missing_cols = list(set(self.data._x) - set(df.columns))
            # for c in missing_cols:
            #     df[c] = np.nan

            # Ensure that the read file's index matches the data's timesteps
            mismatch = df.index.difference(d._dt.index)
            if len(mismatch) > 0:
                e = exceptions.ModelError
                entries = mismatch.tolist()
                raise e('File has invalid index. Ensure that it has the same '
                        'date range and resolution as set_t.csv: {}.\n\n'
                        'Invalid entries: {}'.format(filename, entries))
            return df

        d = self.data
        self.debug.data_sources = utils.AttrDict()
        # Data for storage initialization parameter
        d.s_init = pd.DataFrame(index=d._x, columns=d._y_pc)
        for y in d.s_init.columns:
            for x in d.s_init.index:
                d.s_init.at[x, y] = self.get_option(y + '.constraints.s_init',
                                                    x=x)
        # Parameters that may be defined over (x, t) for a given technology y
        d.params = ['r', 'e_eff']
        d._y_def_r = set()
        d._y_def_e_eff = set()
        # TODO could allow params in d.params to be defined only over
        # x instead of either static or over (x, t) via CSV!
        for param in d.params:
            d[param] = utils.AttrDict()
            for y in d._y:
                d[param][y] = pd.DataFrame(np.nan, index=d._dt.index,
                                           columns=d._x)
                # TODO this whole process could be refactored for efficiency
                # to read files only once,
                # create a dict of files: {'f1.csv': ['x1', 'x2'],
                #                          'f2.csv': ['x3'],
                #                          'model_config': ['x4, x5']}
                for x in d._x:
                    # If this y is actually not defined at this x,
                    # and is also not a transmission tech,
                    # continue (but set the param to 0 first)
                    # TODO this is a bit of a hack -- e.g. the extra check
                    # for transmission tech is necessary because we set
                    # e_eff to 0 for all transmission (as transmission techs
                    # don't show up in the config_model.locations[x].techs)
                    # Keep an eye out in case this causes other problems
                    if (y not in self.config_model.locations[x].techs
                            and y not in d._y_transmission):
                        d[param][y].loc[:, x] = 0
                        continue
                    option = self.get_option(y + '.constraints.' + param, x=x)
                    if (isinstance(option, str)
                            and not option.startswith('file')):
                        e = exceptions.ModelError
                        raise e('Invalid value for `{}.{}.{}`:'
                                ' `{}`'.format(param, y, x, option))
                    k = param + '.' + y + '.' + x
                    if (isinstance(option, str)
                            and option.startswith('file')):
                        if param == 'r':
                            d._y_def_r.add(y)
                        elif param == 'e_eff':
                            d._y_def_e_eff.add(y)
                        try:
                            # Parse 'file=filename' option
                            f = option.split('=')[1]
                        except IndexError:
                            # If set to just 'file', set filename with y and
                            # param, e.g. 'csp_r_eff.csv'
                            f = y + '_' + param + '.csv'
                        df = _get_option_from_csv(f)
                        # Set up x_map if that option has been set
                        try:
                            x_map = self.get_option(y + '.x_map', x=x)
                        except exceptions.OptionNotSetError:
                            x_map = None
                        # If x_map is available, remap the current col
                        if x_map:
                            # TODO this is a hack and will take up a lot
                            # of memory due to data duplication in case
                            # of a lot of mappings pointing to the same
                            # column in the data
                            # Format is <name in model config>:<name in data>
                            x_map_dict = {i.split(':')[0].strip():
                                          i.split(':')[1].strip()
                                          for i in x_map.split(',')}
                            x_map_str = 'x_map: \'{}\''.format(x_map)
                            # Get the mapping for this x from x_map
                            # NB not removing old columns in case
                            # those are also used somewhere!
                            try:
                                x_m = x_map_dict[x]
                            except KeyError:
                                e = exceptions.ModelError
                                raise e('x_map defined but does not map '
                                        'location defined in model config: '
                                        '{}, with {}'.format(x, x_map_str))
                            if x_m not in df.columns:
                                e = exceptions.ModelError
                                raise e('Trying to map to to a column not '
                                        'contained in data: {}, for region '
                                        '{}, with {}'
                                        .format(x_m, x, x_map_str))
                            df[x] = df[x_m]
                        try:
                            d[param][y].loc[:, x] = df[x]
                            self.debug.data_sources.set_key(k, 'file:' + f)
                        except KeyError:
                            # If could not be read from file, set it to zero
                            d[param][y].loc[:, x] = 0
                            # Depending on whether or not the tech is allowed
                            # at this location, set _NA_ for the data source,
                            # or raise an error
                            if self.data.locations.at[x, y] == 0:
                                self.debug.data_sources.set_key(k, '_NA_')
                            else:
                                w = exceptions.ModelWarning
                                message = ('Could not load data for {}, '
                                           'with given option: '
                                           '{}'.format(k, option))
                                warnings.warn(message, w)
                                v = 'file:_NOT_FOUND_'
                                self.debug.data_sources.set_key(k, v)
                    else:
                        d[param][y].loc[:, x] = option
                        self.debug.data_sources.set_key(k, 'model_config')
                        if (param == 'r' and option != float('inf')):
                            d._y_def_r.add(y)
                    # Convert power to energy for r, if necessary
                    if param == 'r':
                        r_unit = self.get_option(y + '.constraints.r_unit')
                        if r_unit == 'power':
                            r_scale = d.time_res_data
                            d[param][y].loc[:, x] = (d[param][y].loc[:, x]
                                                     * r_scale)
                    # Scale r to a given maximum if necessary
                    scale = self.get_option(y + '.constraints.r_scale_to_peak',
                                            x=x)
                    if param == 'r' and scale:
                        scaled = self.scale_to_peak(d[param][y][x], scale)
                        d[param][y].loc[:, x] = scaled
        ds = self.debug.data_sources
        missing_data = [kk for kk in ds.keys_nested()
                        if ds.get_key(kk) == 'file:_NOT_FOUND_']
        if len(missing_data) > 0:
            message = ('The following parameter values could not be read '
                       'from file. They were automatically set to `0`: '
                       + ', '.join(missing_data))
            warnings.warn(message, exceptions.ModelWarning)
        # Finally, check data consistency. For now, demand must be <= 0,
        # and supply >=0, at all times.
        # FIXME update these checks on implementing conditional param updates.
        for y in d._y_def_r:
            base_tech = self.get_parent(y)
            # Check each column:
            for c in d.r[y].columns:
                series = d.r[y][c]
                err_suffix = 'for tech: {}, at location: {}'.format(y, c)
                if base_tech == 'demand':
                    err = 'Demand resource must be <=0, ' + err_suffix
                    assert (series <= 0).all(), err
                elif base_tech == 'supply':
                    err = 'Supply resource must be >=0, ' + err_suffix
                    assert (series >= 0).all(), err

    def _get_t_max_demand(self):
        t_max_demands = utils.AttrDict()
        for c in self.data._c:
            ys = [y for y in self.data._y
                  if self.get_option(y + '.carrier') == c]
            r_carrier = pd.Panel(self.data.r.as_dict()).loc[ys].sum(axis=2)
            t_max_demand = r_carrier[r_carrier < 0].sum(axis=1).idxmin()
            # Adjust for reduced resolution, only if t_max_demand not 0 anyway
            if t_max_demand != 0:
                t_max_demand = max([t for t in self.data._dt.index
                                    if t <= t_max_demand])
            t_max_demands[c] = t_max_demand
        return t_max_demands

    def add_constraint(self, constraint, *args, **kwargs):
        try:
            constraint(self, *args, **kwargs)
        # If there is an error in a constraint, make sure to also get
        # the index where the error happened and pass that along
        except ValueError as e:
            index = inspect.trace()[-1][0].f_locals['index']
            index_string = ', at index: {}'.format(index)
            if not e.args:
                e.args = ('',)
            e.args = (e.args[0] + index_string,) + e.args[1:]
            # Also log it because that is what Pyomo does, and want to ensure
            # that the log entry contains the info we added
            logging.error('Error generating constraint' + index_string)
            raise

    def _param_populator(self, src, t_offset=0):
        """Returns a `getter` function that returns either
        (x, t)-specific values for parameters that define such, or
        always the same static value if only a static value is given.

        """
        def getter(m, y, x, t):
            if isinstance(src[y], pd.core.frame.DataFrame):
                return float(src[y].loc[t + t_offset, x])
            else:
                return float(src[y])
        return getter

    def update_parameters(self, t_offset):
        mi = self.instance
        d = self.data

        for param in d.params:
            initializer = self._param_populator(d[param], t_offset)
            y_set = self.param_sets[param]
            param_object = getattr(mi, param)
            for y in y_set:
                for x in mi.x:
                    for t in mi.t:
                        param_object[y, x, t] = initializer(mi, y, x, t)

        s_init_initializer = lambda m, y, x: float(d.s_init.at[x, y])
        for y in mi.y_pc:
            for x in mi.x:
                mi.s_init[y, x] = s_init_initializer(mi, y, x)

    def _set_t_end(self):
        # t_end is the timestep previous to t_start + horizon,
        # because the .loc[start:end] slice includes the end
        try:
            self.t_end = self.prev(int(self.t_start
                                   + self.config_model.opmode.horizon
                                   / self.data.time_res_data))
        except KeyError:
            # If t_end is beyond last timestep, cap it to last one, and
            # log the occurance
            t_bound = self.data._dt.index[-1]
            msg = 'Capping t_end to {}'.format(t_bound)
            logging.debug(msg)
            self.t_end = t_bound
        # print('\n\n***\n{}-{}\n***\n\n'.format(self.t_start, self.t_end))

    def generate_model(self, t_start=None):
        """
        Generate the model and store it under the property `m`.

        Args:
            t_start : if self.mode == 'operate', this must be specified,
            but that is done automatically via solve_iterative() when
            calling run()

        """
        #
        # Setup
        #
        m = po.ConcreteModel()
        o = self.config_model
        d = self.data
        self.m = m
        self.t_start = t_start
        self.t_max_demand = self._get_t_max_demand()

        #
        # Sets
        #

        # Time steps
        if self.mode == 'plan':
            m.t = po.Set(initialize=d._dt.index, ordered=True)
        elif self.mode == 'operate':
            self._set_t_end()
            m.t = po.Set(initialize=d._dt.loc[self.t_start:self.t_end].index,
                         ordered=True)
        # Carriers
        m.c = po.Set(initialize=d._c, ordered=True)
        # Locations
        m.x = po.Set(initialize=d._x, ordered=True)
        # Cost classes
        m.k = po.Set(initialize=d._k, ordered=True)
        #
        # Technologies and various subsets of technologies
        #
        m.y = po.Set(initialize=d._y, ordered=True)
        # Production technologies
        m.y_prod = po.Set(initialize=d._y_prod, within=m.y, ordered=True)
        # Production technologies
        m.y_con = po.Set(initialize=d._y_con, within=m.y, ordered=True)
        # Production/consumption technologies
        m.y_pc = po.Set(initialize=d._y_pc, within=m.y, ordered=True)
        # Transmission technologies
        m.y_trans = po.Set(initialize=d._y_transmission, within=m.y,
                           ordered=True)
        # Conversion technologies
        m.y_conv = po.Set(initialize=d._y_conversion, within=m.y,
                          ordered=True)
        # Technologies with specified `r`
        m.y_def_r = po.Set(initialize=d._y_def_r, within=m.y)
        # Technologies with specified `e_eff`
        m.y_def_e_eff = po.Set(initialize=d._y_def_e_eff, within=m.y)
        # Technologies that allow `rb`
        m.y_rb = po.Set(initialize=d._y_rb, within=m.y)
        # Technologies with parasitics
        m.y_p = po.Set(initialize=d._y_p, within=m.y)
        # Technologies without parasitics
        m.y_np = po.Set(initialize=set(d._y) - set(d._y_p), within=m.y)

        #
        # Parameters
        #

        self.param_sets = {'r': m.y_def_r,
                           'a': m.y_def_r,
                           'e_eff': m.y_def_e_eff}
        for param in d.params:
            initializer = self._param_populator(d[param])
            y = self.param_sets[param]
            setattr(m, param, po.Param(y, m.x, m.t, initialize=initializer,
                                       mutable=True))

        s_init_initializer = lambda m, y, x: float(d.s_init.at[x, y])
        m.s_init = po.Param(m.y_pc, m.x, initialize=s_init_initializer,
                            mutable=True)

        #
        # Variables and constraints
        #
        # 1. Required
        required = [constraints.base.node_resource,
                    constraints.base.node_energy_balance,
                    constraints.base.node_constraints_build,
                    constraints.base.node_constraints_operational,
                    constraints.base.node_constraints_transmission,
                    constraints.base.node_parasitics,
                    constraints.base.node_costs,
                    constraints.base.model_constraints]
        for c in required:
            self.add_constraint(c)

        if self.mode == 'plan':
            self.add_constraint(constraints.planning.system_margin)

        # 2. Optional
        if o.get_key('constraints', default=False):
            for c in o.constraints:
                self.add_constraint(_load_function(c))

        # 3. Objective function
        default_obj = 'constraints.objective.objective_cost_minimization'
        objective = o.get_key('objective', default=default_obj)
        self.add_constraint(_load_function(objective))

    def _log_time(self):
        self.runtime = int(time.time() - self.start_time)
        logging.info('Runtime: ' + str(self.runtime) + ' secs')

    def run(self, iterative_warmstart=True):
        """
        Instantiate and solve the model

        """
        o = self.config_model
        d = self.data
        cr = self.config_run
        self.start_time = time.time()
        if self.mode == 'plan':
            self.generate_model()  # Generated model goes to self.m
            self.solve()
            self.load_solution()
        elif self.mode == 'operate':
            assert len(self.data.time_res_series.unique()) == 1, \
                'Operational mode only works with uniform time step lengths.'
            assert (d.time_res_static <= o.opmode.horizon and
                    d.time_res_static <= o.opmode.window), \
                'Timestep length must be smaller than horizon and window.'
            # solve_iterative() generates, solves, and loads the solution
            self.solve_iterative(iterative_warmstart)
        else:
            e = exceptions.ModelError
            raise e('Invalid model mode: `{}`'.format(self.mode))
        self._log_time()
        if cr.get_key('output.save', default=False) is True:
            output_format = cr.get_key('output.format', default=['hdf'])
            if not isinstance(output_format, list):
                output_format = [output_format]
            for fmt in output_format:
                self.save_solution(fmt)
            save_constr = cr.get_key('output.save_constraints', default=False)
            if save_constr:
                options = cr.get_key('output.save_constraints_options',
                                     default={})
                output.generate_constraints(self.solution,
                                            output_path=save_constr,
                                            **options)

    def solve(self, warmstart=False):
        """
        Args:
            warmstart : (default False) re-solve an updated model
                        instance

        Returns: None

        """
        m = self.m
        cr = self.config_run
        if not warmstart:
            self.instance = m.create()
            solver_io = cr.get_key('solver_io', default=False)
            if solver_io:
                self.opt = popt.SolverFactory(cr.solver, solver_io=solver_io)
            else:
                self.opt = popt.SolverFactory(cr.solver)
            # Set solver options from run_settings file, if it exists
            try:
                for k in cr.solver_options.keys_nested():
                    self.opt.options[k] = cr.solver_options.get_key(k)
            except KeyError:
                pass
            if cr.get_key('debug.symbolic_solver_labels', default=False):
                self.opt.symbolic_solver_labels = True
        if cr.get_key('debug.keep_temp_files', default=False):
            self.opt.keepfiles = True
            if self.mode == 'plan':
                logdir = os.path.join('Logs', self.run_id)
            elif self.mode == 'operate':
                logdir = os.path.join('Logs', self.run_id
                                      + '_' + str(self.t_start))
            if (cr.get_key('debug.overwrite_temp_files', default=False)
                    and os.path.exists(logdir)):
                shutil.rmtree(logdir)
            os.makedirs(logdir)
            TempfileManager.tempdir = logdir
        # Always preprocess instance, for both cold start and warm start
        self.instance.preprocess()

        def _solve(warmstart):
            if warmstart:
                results = self.opt.solve(self.instance, warmstart=True,
                                         tee=True)
            else:
                results = self.opt.solve(self.instance, tee=True)
            return results

        if self.verbose:
            t = datetime.datetime.now().strftime(self.time_format)
            print('\nModel preprocessing complete at {}'.format(t))

        if cr.get_key('debug.echo_solver_log', default=False):
            self.results = _solve(warmstart)
        else:
            # Silencing output by redirecting stdout and stderr
            with utils.capture_output() as self.pyomo_output:
                self.results = _solve(warmstart)
        self.load_results()

    def process_solution(self):
        # Add levelized cost
        self.solution['levelized_cost'] = self.get_levelized_cost()
        # Add capacity factor
        self.solution['capacity_factor'] = self.get_capacity_factor()
        # Add metadata
        self.solution['metadata'] = self.get_metadata()
        # Add summary
        self.solution['summary'] = self.get_summary()
        # Add shares
        self.solution['shares'] = self.get_shares()
        # Add time resolution, and give it a nicer index
        time_res = self.data.time_res_series.copy()
        time_res.index = self.solution.node.major_axis
        self.solution['time_res'] = time_res
        # Add model and run config
        self.solution['config_run'] = self.config_run
        self.solution['config_model'] = self.config_model

    def load_solution(self):
        sol = {'node': self.get_node_variables(),
               'totals': self.get_totals(),
               'costs': self.get_costs(),
               'parameters': self.get_node_parameters()}
        self.solution = utils.AttrDict(sol)
        self.process_solution()

    def get_var(self, var, dims=None):
        """
        Return output for variable `var` as a series, dataframe or panel

        Args:
            var : variable name as string, e.g. 'es_prod'

            dims : list of indices as strings, e.g. ('y', 'x', 't');
            if not given, they are auto-detected

        """
        # FIXME: even if `dims` were given where `t` is not in the last
        # position, the code below assumes that `t` is in the last position,
        # and won't work otherwise. All model variables are formulated
        # so that `t` is always last, so this should not be a problem.
        m = self.m
        try:
            var = getattr(m, var)
        except AttributeError:
            raise exceptions.ModelError('Variable {} inexistent.'.format(var))
        # Get dims
        if not dims:
            dims = [i.name for i in var.index_set().set_tuple]
        result = pd.DataFrame.from_dict(var.get_values(), orient='index')
        if result.empty:
            raise exceptions.ModelError('Variable {} has no data.'.format(var))
        result.index = pd.MultiIndex.from_tuples(result.index, names=dims)
        result = result[0]  # Get the only column in the dataframe
        # Unstack and sort by time axis
        if len(dims) == 1:
            result = result.sort_index()
        else:
            # if len(dims) is 2, we already have a well-formed panel
            result = result.unstack(level=0)
            if len(dims) == 2:
                result = result.sort_index()
            if len(dims) == 3:
                result = result.to_panel()
                result = result.transpose(0, 2, 1)  # Flip time into major_axis
                result = result.sort_index(1)
            if len(dims) == 4:
                # Panel4D needs manual treatment until pandas supports to_panel
                # for anything else but the regular 3d panel
                p = {}
                for i in result.columns:
                    p[i] = result[i].unstack(0).to_panel().transpose(0, 2, 1)
                result = pd.Panel4D(p)
                result = result.sort_index(1)
        # Nicify time axis
        if 't' in dims:
            t = getattr(m, 't')
            if self.t_start is None:
                new_index = self.data._dt.loc[t.first():t.last()].tolist()
            else:
                new_index = self.data._dt.loc[self.t_start:self.t_end].tolist()
            if len(dims) <= 2:
                result.index = new_index
            else:
                result.major_axis = new_index
        return result

    def get_ec(self, what='prod'):
        es = self.get_var('es_' + what)
        try:
            ec = self.get_var('ec_' + what)
        except exceptions.ModelError:  # ec has no data
            # Skip all the rest and return es straight away
            return es
        # For those techs that have an `ec`, replace their `es` with `ec`,
        # for the others, `ec` is `es`, so no change needed
        for carrier in ec.labels:
            for tech in ec.items:
                es[carrier][tech] = ec[carrier][tech]
        return es  # the variable is called es, but the thing is now ec

    def get_ec_sum(self):
        prod = self.get_ec('prod')
        con = self.get_ec('con')
        ec = {}
        for i in prod.labels:
            ec[i] = (prod[i].to_frame().add(con[i].to_frame(),
                                            fill_value=0)).to_panel()
        return pd.Panel4D(ec)

    def get_node_variables(self):
        detail = ['s', 'rs']
        p = pd.Panel4D({v: self.get_var(v) for v in detail})
        try:
            p['rbs'] = self.get_var('rbs')
        except exceptions.ModelError:
            # `rbs` doesn't exist in the model or exists without data
            p['rbs'] = 0
        # Add 'e:c' items for each c in carrier
        temp = self.get_ec_sum()
        for c in self.data._c:
            p['e' + ':' + c] = temp.loc[c, :, :, :]
        return p

    def get_node_parameters(self, built_only=False):
        """If built_only is True, disregard locations where e_cap==0"""
        detail = ['s_cap', 'r_cap', 'r_area', 'e_cap', 'e_cap_net']
        result = pd.Panel({v: self.get_var(v) for v in detail})
        try:
            result['rb_cap'] = self.get_var('rb_cap')
        except exceptions.ModelError:
            result['rb_cap'] = 0
        if built_only:
            result = result.to_frame()
            result = result[result.e_cap > 0].dropna(axis=1).to_panel()
        return result

    def get_costs(self, t_subset=None):
        """Get costs."""
        def _factor(var, t_subset):
            """Return the fraction of var within t_subset, used to
               calculate the correct fraction of operational costs."""
            if var.ndim == 4:  # Make Panel4D into Panel by summing carriers
                var = var.sum(0)
            return (var.iloc[:, t_subset, :].sum(1)
                    / var.sum(1)).fillna(0)

        if t_subset is None:
            t_subset = slice(None)

        T = lambda x: pd.Panel.transpose(x, 'major_axis',
                                         'minor_axis', 'items')

        len_adjust = (len(self.data.time_res_series.iloc[t_subset])
                      / len(self.m.t))

        cost = T(self.get_var('cost_con').add(self.get_var('cost_op_fixed')))
        # Adjust for the fact that fixed costs accrue over a smaller length
        # of time as per len_adjust
        cost = cost * len_adjust
        cost_var = T(self.get_var('cost_op_var'))
        cost_fuel = T(self.get_var('cost_op_fuel'))
        cost_rb = T(self.get_var('cost_op_rb'))
        # Adjust for the fact that var and fuel costs are only accrued over
        # the t_subset period
        if t_subset:
            es_prod = self.get_var('es_prod')
            # Broadcast multiplication along items (i.e. cost classes)
            cost_var = cost_var.multiply(_factor(es_prod, t_subset),
                                         axis='items')
            rs = self.get_var('rs')
            cost_fuel = cost_fuel.multiply(_factor(rs, t_subset),
                                           axis='items')
            try:
                rbs = self.get_var('rbs')
                cost_rb = cost_rb.multiply(_factor(rbs, t_subset),
                                           axis='items')
            except exceptions.ModelError:
                pass  # If rbs doesn't exist in the data, ModelError is raised,
                # and we simply move on...
        cost = cost.add(cost_var).add(cost_fuel).add(cost_rb)
        return cost

    def get_totals(self, t_subset=None):
        """Get total produced and consumed per technology and location."""
        if t_subset is None:
            t_subset = slice(None)
        p = pd.Panel4D({'ec_' + i: self.get_ec(i)
                        .iloc[:, :, t_subset, :]
                        .sum(axis='major')
                        for i in ['prod', 'con']})
        for i in ['es_prod', 'es_con']:
            p[i] = self.get_var(i).iloc[:, :, t_subset, :].sum(axis='major')
        p = p.transpose('items', 'labels', 'minor_axis', 'major_axis')
        return p

    def get_levelized_cost(self):
        """
        Get levelized costs.

        NB: Only production, not consumption, is used in calculations.

        """
        sol = self.solution
        p4d = {}
        for cost in self.data._k:
            p = {}
            for carrier in self.data._c:
                # Levelized cost of electricity (LCOE)
                lc = sol.costs[cost] / sol.totals[carrier].ec_prod
                lc[np.isinf(lc)] = 0
                lc.loc['total', :] = (sol.costs[cost].sum(0)
                                      / sol.totals[carrier].ec_prod.sum(0))
                p[carrier] = lc
            p4d[cost] = pd.Panel(p)
        return pd.Panel4D(p4d)

    def get_capacity_factor(self):
        """
        Get capacity factor.

        NB: Only production, not consumption, is used in calculations.

        """
        m = self.m
        sol = self.solution
        time_res = self.data.time_res_series
        e_cap = sol.parameters['e_cap_net']
        cfs = {}
        for carrier in sol.totals.labels:
            try:  # Try loading time_res_sum from operational mode
                time_res_sum = self.data.time_res_sum
            except KeyError:
                time_res_sum = sum(time_res.at[t] for t in m.t)
            cf = sol.totals[carrier].ec_prod / (e_cap * time_res_sum)
            cf.loc['total', :] = (sol.totals[carrier].ec_prod.sum(0)
                                  / (e_cap.sum(0) * time_res_sum))
            cf = cf.fillna(0)
            cfs[carrier] = cf
        return pd.Panel(cfs)

    def get_metadata(self):
        df = pd.DataFrame(index=self.data._y)
        df.loc[:, 'type'] = df.index.map(lambda y: self.get_parent(y))
        df.loc[:, 'name'] = df.index.map(lambda y: self.get_name(y))
        df.loc[:, 'carrier'] = df.index.map(lambda y: self.get_carrier(y))
        get_src_c = lambda y: self.get_source_carrier(y)
        df.loc[:, 'source_carrier'] = df.index.map(get_src_c)
        df.loc[:, 'stack_weight'] = df.index.map(lambda y: self.get_weight(y))
        df.loc[:, 'color'] = df.index.map(lambda y: self.get_color(y))
        return df

    def get_summary(self, sort_by='e_cap', carrier='power'):
        sol = self.solution
        # Capacity factor per carrier
        df = pd.DataFrame({'cf': sol.capacity_factor.loc[carrier, 'total', :]})
        # Costs per carrier
        for k in sorted(sol.levelized_cost.labels):
            # .loc[cost_class, carrier, location, tech]
            df['cost_' + k] = sol.levelized_cost.loc[k, carrier, 'total', :]
        # Add totals per carrier
        df['e_prod'] = sol.totals.loc[carrier, 'ec_prod', :, :].sum(0)
        df['e_con'] = sol.totals.loc[carrier, 'ec_con', :, :].sum(0)
        # Add other carrier-independent stuff
        df['e_cap'] = sol.parameters['e_cap'].sum()
        df['r_area'] = sol.parameters['r_area'].sum()
        return df.sort(columns=sort_by, ascending=False)

    def get_shares(self):
        from . import analysis
        ggm = self.get_group_members
        s = pd.Series({k: '|'.join(ggm(k, head_nodes_only=True))
                      for k in self.config_model.techs
                      if ggm(k, head_nodes_only=True) != []
                      and ggm(k, head_nodes_only=True) is not None})

        df = pd.DataFrame(s, columns=['members'])
        gg = lambda y: self.get_option(y + '.group',
                                       default='defaults.group',
                                       ignore_inheritance=True)
        df['group'] = df.index.map(gg)
        df['type'] = df.index.map(self.get_parent)

        for var in ['e_prod', 'e_con', 'e_cap']:
            for index, row in df.iterrows():
                group_members = row['members'].split('|')
                group_type = row['type']
                share = analysis.get_group_share(self.solution, group_members,
                                                 group_type, var=var)
                df.at[index, var] = share
        return df

    def load_solution_iterative(self, node_vars, total_vars, cost_vars):
        def _panel_sum(panels):
            # Sums a list of 3d or 4d panels
            result = panels[0]  # Take first item
            for i in range(1, len(panels)):  # Start at 1 as we dealt with 0
                result = result.add(panels[i])
            return result
        totals = _panel_sum(total_vars)
        costs = _panel_sum(cost_vars)
        sol = {'node': pd.concat(node_vars, axis=2),
               'totals': totals,
               'costs': costs,
               'parameters': self.get_node_parameters()}
        self.solution = utils.AttrDict(sol)
        self.process_solution()

    def solve_iterative(self, iterative_warmstart=True):
        """
        Solve iterative by updating model parameters.

        By default, on optimizations subsequent to the first one,
        warmstart is used to speed up the model generation process.

        Returns None on success, storing results under self.solution

        """
        o = self.config_model
        d = self.data
        time_res = d.time_res_series
        window_adj = int(o.opmode.window / d.time_res_data)
        steps = [t for t in d._dt.index if (t % window_adj) == 0]
        # Remove the last step - since we look forward at each step,
        # it would take us beyond actually existing data
        steps = steps[:-1]
        node_vars = []
        total_vars = []
        cost_vars = []
        d.time_res_sum = 0
        self.generate_model(t_start=steps[0])
        for index, step in enumerate(steps):
            if index == 0:
                self.solve(warmstart=False)
            else:
                self.t_start = step
                self._set_t_end()
                # Note: we don't update the timestep set, so it keeps the
                # values it got on first construction. Instead,
                # we use an offset when updating parameter data so that
                # the correct values are read into the "incorrect" timesteps.
                self.update_parameters(t_offset=step - steps[0])
                self.solve(warmstart=iterative_warmstart)
            self.load_results()
            # Gather relevant model results over decision interval, so
            # we only grab [0:window/time_res_static] steps, where
            # window/time_res_static will be an iloc index
            if index == (len(steps) - 1):
                # Final iteration saves data from entire horizon
                stepsize = int(o.opmode.horizon / d.time_res_static)
            else:
                # Non-final iterations only save data from window
                stepsize = int(o.opmode.window / d.time_res_static)
            # Get node variables
            panel4d = self.get_node_variables()
            node_vars.append(panel4d.iloc[:, :, 0:stepsize, :])
            # Get totals
            totals = self.get_totals(t_subset=slice(0, stepsize))
            total_vars.append(totals)
            # Get costs
            cost = self.get_costs(t_subset=slice(0, stepsize))
            cost_vars.append(cost)
            timesteps = [time_res.at[t] for t in self.m.t][0:stepsize]
            d.time_res_sum += sum(timesteps)
            # Save state of storage for carry over to next iteration
            s = self.get_var('s')
            # Convert from timestep length to absolute index
            storage_state_index = stepsize - 1
            assert (isinstance(storage_state_index, int) or
                    storage_state_index.is_integer())
            storage_state_index = int(storage_state_index)
            d.s_init = s.iloc[:, storage_state_index, :]
        self.load_solution_iterative(node_vars, total_vars, cost_vars)

    def load_results(self):
        """Load results into model instance for access via model variables."""
        r = self.instance.load(self.results)
        if r is False:
            logging.critical(self.results.Problem)
            logging.critical(self.results.Solver)
            w = exceptions.ModelWarning
            warnings.warn('Could not load results into model instance.', w)

    def save_solution(self, how):
        """Save model solution. ``how`` can be 'hdf' or 'csv'

        CSV is supported for legacy purposes but usually, the
        HDF option should be used.

        """
        # Create output dir, but ignore if it already exists
        try:
            os.makedirs(self.config_run.output.path)
        except OSError:  # Hoping this isn't raised for more serious stuff
            pass
        if how == 'hdf':
            self._save_hdf()
        elif how == 'csv':
            self._save_csv()
        else:
            raise ValueError('Unsupported value for `how`: {}'.format(how))

    def _save_hdf(self):
        """
        Save solution as HDF5 to the file ``solution.hdf`` in
        ``self.config_run.output.path``

        """
        sol = self.solution
        store_file = os.path.join(self.config_run.output.path, 'solution.hdf')
        # Raise error if file exists already, to make sure we don't destroy
        # existing data
        if os.path.exists(store_file):
            i = 0
            alt_file = os.path.join(self.config_run.output.path,
                                    'solution_{}.hdf')
            while os.path.exists(alt_file.format(i)):
                i += 1
            alt_file = alt_file.format(i)  # Now "pick" the first free filename
            message = ('File `{}` exists, '
                       'using `{}` instead.'.format(store_file, alt_file))
            logging.warning(message)
            store_file = alt_file
        # Use zlib compression (HDF standard)
        # Using mode 'w' means existing file will be overwritten
        store = pd.HDFStore(store_file, mode='w',
                            complevel=3, complib='zlib')
        # Now save solution -- except for config_model and config_run, since
        # they are AttrDicts so are saved separately below
        for key in [k for k in sol if k not in ['config_model', 'config_run']]:
            # Use .append instead of .add for Panel4D compatibility
            store.append(key, sol[key])
        # Save r and e_eff time series alongside the solution
        for k in self.data._y_def_r:
            store.append('data/r/' + k, self.data.r[k])
        for k in self.data._y_def_e_eff:
            store.append('data/e_eff/' + k, self.data.e_eff[k])
        # Now, save config_model and config_run as YAML strings
        config = pd.Series({key: sol[key].to_yaml() for key in ['config_model', 'config_run']})
        store.append('config', config)
        store.close()
        # Return the path we used
        return store_file

    def _save_csv(self):
        """Save solution as CSV files to ``self.config_run.output.path``"""
        sol = self.solution
        output_files = {'node_parameters.csv': sol.parameters.to_frame(),
                        'costs.csv': sol.costs.to_frame(),
                        'capacity_factor.csv': sol.capacity_factor.to_frame(),
                        'metadata.csv': sol.metadata,
                        'summary.csv': sol.summary,
                        'shares.csv': sol.shares,
                        'time_res.csv': sol.time_res}
        for var in sol.node.labels:
            k = 'node_variables_{}.csv'.format(var.replace(':', '_'))
            v = sol.node[var].to_frame()
            output_files[k] = v
        for c in self.data._c:
            k = 'totals_{}.csv'.format(c)
            output_files[k] = sol.totals[c].to_frame()
        for cost_class in self.data._k:
            k = 'levelized_cost_{}.csv'.format(cost_class)
            output_files[k] = sol.levelized_cost[cost_class].to_frame()
        # Write all files to output dir
        for k, v in output_files.items():
            v.to_csv(os.path.join(self.config_run.output.path, k))
        # Save r and e_eff time series alongside the solution
        for k in self.data._y_def_r:
            pth = os.path.join(self.config_run.output.path, 'data_r_' + k)
            self.data.r[k].to_csv(pth)
        for k in self.data._y_def_e_eff:
            pth = os.path.join(self.config_run.output.path, 'data_e_eff_' + k)
            self.data.e_eff[k].to_csv(pth)
        # Also save model and run configuration
        self.config_run.to_yaml(os.path.join(self.config_run.output.path,
                                             'config_run.yaml'))
        self.config_model.to_yaml(os.path.join(self.config_run.output.path,
                                               'config_model.yaml'))



"""
Copyright (C) 2013-2015 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

planning.py
~~~~~~~~~~~

Planning constraints.

"""

import pyomo.core as po


def system_margin(model):
    """

    """
    m = model.m
    time_res = model.data.time_res_series

    def carrier(y):
        return model.get_option(y + '.carrier')

    # Constraint rules
    def c_system_margin_rule(m, c):
        # If no margin defined for a carrier, use 0 (i.e. no margin)
        margin = model.config_model.system_margin.get_key(c, default=0)
        if margin:
            t = model.t_max_demand[c]
            return (sum(m.es_prod[c, y, x, t] for y in m.y for x in m.x)
                    * (1 + margin)
                    <= time_res.at[t]
                    * sum((m.e_cap[y, x] / model.get_eff_ref('e', y, x))
                          for y in m.y if carrier(y) == c
                          for x in m.x))
        else:
            return po.Constraint.NoConstraint

    # Constraints
    m.c_system_margin = po.Constraint(m.c, rule=c_system_margin_rule)



import pytest
import tempfile

from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver


class TestModel:
    @pytest.fixture(scope='module')
    def model(self):
        locations = """
            locations:
                1:
                    level: 0
                    within:
                    techs: ['ccgt', 'test_storage', 'demand_electricity',
                            'unmet_demand_electricity']
                    override:
                        test_storage:
                            constraints:
                                e_cap_max: 0.5
                                s_init: 0
                        ccgt:
                            constraints:
                                e_cap_max: 9.5
                        demand_electricity:
                            x_map: '1: demand'
                            constraints:
                                r: file=demand-sin_r.csv
            links:
        """
        config_run = """
            mode: plan
            model: [{techs}, {locations}]
            subset_t: ['2005-01-01', '2005-01-03']
        """
        with tempfile.NamedTemporaryFile() as f:
            f.write(locations.encode('utf-8'))
            f.read()
            model = common.simple_model(config_run=config_run,
                                        config_locations=f.name,
                                        override=AttrDict({'solver': solver}))
        model.run()
        return model

    def test_model_solves(self, model):
         assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_balanced(self, model):
        df = model.solution.node
        assert df.loc['e:power', 'ccgt', :, :].iloc[0, :].sum() == 8.0
        assert_almost_equal(df.loc['e:power', 'ccgt', :, :].sum(1).mean(),
                            7.62, tolerance=0.01)
